from viewflow import this

from ..base import Node, Edge
from ..activation import Activation
from . import mixins


class IfActivation(Activation):
    """Conditionally activate one of outgoing nodes."""

    _condition_result = None

    @Activation.status.super()
    def activate(self):
        self._condition_result = self.flow_task.condition(self)

    @Activation.status.super()
    def create_next(self):
        next_node = (
            self.flow_task._on_true
            if self._condition_result
            else self.flow_task._on_false
        )
        if next_node:
            yield next_node._create(self, self.task.token)


class If(Node):
    """If gateway, activate one of outgoing node."""

    activation_class = IfActivation

    task_type = "EXCLUSIVE_GATEWAY"

    shape = {
        "width": 50,
        "height": 50,
        "svg": """
            <path class="gateway" d="M25,0L50,25L25,50L0,25L25,0"/>
            <text class="gateway-marker" font-size="16px" x="25" y="31">X</text>
        """,
    }

    bpmn_element = "exclusiveGateway"

    def __init__(self, cond, **kwargs):
        super().__init__(**kwargs)
        self._condition = cond
        self._on_true = None
        self._on_false = None

    def _outgoing(self):
        yield Edge(src=self, dst=self._on_true, edge_class="cond_true")
        yield Edge(src=self, dst=self._on_false, edge_class="cond_false")

    def _resolve(self, instance):
        super()._resolve(instance)
        self._on_true = this.resolve(instance, self._on_true)
        self._on_false = this.resolve(instance, self._on_false)

    @property
    def condition(self):
        return this.resolve(self.flow_class.instance, self._condition)

    def Then(self, node):
        """Node activated if condition is True."""
        self._on_true = node
        return self

    def Else(self, node):
        """Node activated if condition is False."""
        self._on_false = node
        return self



from django.dispatch import Signal

# providing_args=["process", "task"]
flow_started = Signal()

# providing_args=["process", "task"]
flow_finished = Signal()

# providing_args=["process", "task"]
task_started = Signal()

# providing_args=["process", "task", "exception", "traceback"]
task_failed = Signal()

# providing_args=["process", "task"]
task_finished = Signal()



from django.db import transaction
from django.utils.timezone import now

from ..base import Node
from ..activation import Activation, process_not_cancelled
from ..status import STATUS, PROCESS
from ..signals import task_started, task_finished, flow_finished


class EndActivation(Activation):
    """Activation that finishes the flow process."""

    @Activation.status.transition(
        source=STATUS.DONE, target=STATUS.CANCELED, conditions=[process_not_cancelled]
    )
    def undo(self):
        self.process.finished = None
        self.process.status = PROCESS.NEW
        self.process.save()
        super().undo.original()

    @Activation.status.super()
    def activate(self):
        """
        Finalize the flow. If there is no active task, process marked as finished.
        """
        with transaction.atomic(savepoint=True):
            self.task.started = now()
            task_started.send(
                sender=self.flow_class, process=self.process, task=self.task
            )

            active_tasks_count = (
                self.flow_class.task_class._default_manager.filter(
                    process=self.process, finished__isnull=True
                ).exclude(pk=self.task.pk)
            ).count()

            if active_tasks_count == 0:
                self.process.status = STATUS.DONE
                self.process.finished = now()
                self.process.save()

            task_finished.send(
                sender=self.flow_class, process=self.process, task=self.task
            )
            if active_tasks_count == 0:
                flow_finished.send(
                    sender=self.flow_class, process=self.process, task=self.task
                )

    @Activation.status.super()
    def create_next(self):
        """Do nothing"""
        return []


class End(Node):
    """
    End of the flow.

    If no other parallel activities exists, finishes the whole
    process.
    """

    activation_class = EndActivation

    task_type = "END"

    shape = {
        "width": 50,
        "height": 50,
        "svg": """
            <circle class="event end-event" cx="25" cy="25" r="25"/>
        """,
    }

    bpmn_element = "endEvent"

    def _outgoing(self):
        return iter([])



import copy

from viewflow.fsm import State
from ..activation import Activation, STATUS, has_manage_permission
from ..base import Node
from . import mixins


class ObsoleteActivation(Activation):
    @Activation.status.transition(
        source=State.ANY,
        target=STATUS.CANCELED,
        conditions=[lambda a: a.task.status != STATUS.CANCELED],
        permission=has_manage_permission,
    )
    def cancel(self):
        if self.flow_task._cancel_func is not None:
            self.flow_task._cancel_func(self)
        self.task.save()


class Obsolete(Node):
    """Missing node instance."""

    activation_class = ObsoleteActivation

    task_type = "OBSOLETE"

    shape = {"width": 0, "height": 0, "svg": ""}

    bpmn_element = None

    def __init__(self, cancel_func=None, undo_func=None, **kwargs):
        super().__init__(**kwargs)
        self._undo_func = undo_func
        self._cancel_func = cancel_func

    def _outgoing(self):
        return
        yield

    def create_node(self, name, flow_class):
        """
        Create real node instance for missing entry
        """
        obsolete = copy.copy(self)
        obsolete.name = name
        obsolete.flow_class = flow_class
        return obsolete



from django.utils.timezone import now

from viewflow import this
from viewflow.utils import is_owner
from ..base import Node
from ..activation import Activation, has_manage_permission
from ..status import STATUS
from ..signals import task_started, task_finished
from . import mixins


class ViewActivation(mixins.NextNodeActivationMixin, Activation):
    """View node activation."""

    @classmethod
    def create(cls, flow_task, prev_activation, token):
        """Instantiate and persist new flow task."""
        flow_class = flow_task.flow_class
        task = flow_class.task_class(
            process=prev_activation.process, flow_task=flow_task, token=token
        )

        activation = cls(task)

        # Try to assign permission
        owner_permission = flow_task.calc_owner_permission(activation)
        if owner_permission:
            task.owner_permission = owner_permission
            task.owner_permission_obj = flow_task.calc_owner_permission_obj(activation)

        # Try to assign owner
        owner = flow_task.calc_owner(activation)
        if owner:
            task.owner = owner
            task.status = STATUS.ASSIGNED

        task.save()
        task.previous.add(prev_activation.task)

        if flow_task._on_create is not None:
            flow_task._on_create(activation)

        return activation

    @Activation.status.transition(
        source=STATUS.NEW,
        target=STATUS.ASSIGNED,
        permission=lambda activation, user: activation.flow_task.can_assign(
            user, activation.task
        ),
    )
    def assign(self, user):
        """Assign user to the task."""
        self.task.owner = user.id
        self.task.save()

    @Activation.status.transition(
        source=STATUS.ASSIGNED,
        target=STATUS.NEW,
        permission=lambda activation, user: activation.flow_task.can_unassign(
            user, activation.task
        ),
    )
    def unassign(self):
        """Remove user from the task assignment."""
        self.task.owner = None
        self.task.save()

    @Activation.status.transition(
        source=STATUS.ASSIGNED,
        permission=lambda activation, user: activation.flow_task.can_unassign(
            user, activation.task
        ),
    )
    def reassign(self, user=None):
        """Reassign to another user."""
        if user:
            self.task.owner = user
        self.task.save()

    @Activation.status.super()
    def activate(self):
        """Do nothing on sync call"""

    @Activation.status.transition(
        label="Execute",
        source=STATUS.ASSIGNED,
        target=STATUS.STARTED,
        permission=lambda activation, user: activation.flow_task.can_execute(
            user, activation.task
        ),
    )
    def start(self, request):
        # TODO request.GET['started']
        task_started.send(sender=self.flow_class, process=self.process, task=self.task)
        self.task.started = now()

    @Activation.status.transition(source=STATUS.STARTED, target=STATUS.DONE)
    def complete(self):
        """Complete task and create next."""
        super().complete.original()

    @Activation.status.transition(
        source=STATUS.STARTED,
        permission=lambda activation, user: activation.flow_task.can_execute(
            user, activation.task
        ),
    )
    def execute(self):
        self.complete()
        task_finished.send(sender=self.flow_class, process=self.process, task=self.task)
        self.activate_next()

    @Activation.status.transition(
        source=[STATUS.NEW, STATUS.ASSIGNED],
        target=STATUS.CANCELED,
        permission=has_manage_permission,
    )
    def cancel(self):
        self.task.finished = now()
        self.task.save()

    @Activation.status.super()
    def undo(self):
        # undo
        if self.flow_task._undo_func is not None:
            self.flow_task._undo_func(self)

        super().undo.original()


class View(
    mixins.NextNodeMixin,
    mixins.NodePermissionMixin,
    Node,
):
    """User task."""

    activation_class = ViewActivation

    task_type = "HUMAN"

    shape = {
        "width": 150,
        "height": 100,
        "text-align": "middle",
        "svg": """
            <rect class="task" width="150" height="100" rx="5" ry="5"/>
        """,
    }

    bpmn_element = "userTask"

    def __init__(self, view, undo_func=None, **kwargs):
        super().__init__()
        self._view = view
        self._undo_func = undo_func
        self._on_create = None

    def _resolve(self, instance):
        super()._resolve(instance)
        self._on_create = this.resolve(instance, self._on_create)

    """
    Task assign permissions
    """

    def Assign(self, owner=None, **owner_kwargs):
        """
        Assign task to the User immediately on activation.

        Accepts user lookup kwargs or callable :: Process -> User::

            .Assign(username='employee')
            .Assign(lambda activation: activation.process.created_by)
        """
        if owner:
            self._owner = owner
        else:
            self._owner = owner_kwargs
        return self

    def onCreate(self, ref):
        """
        Call a function when task created::

            class MyFlow(Flow):
                approve = flow.View(...).OnCreate(this.on_approve_created)

                def on_approve_created(self, activation):
                    if activation.task.owner:
                        send_mail(
                            'View task assigned to you','Here is the message.',
                            'from@example.com', [activation.task.owner.email]
                        )
        """
        self._on_create = ref
        return self

    def calc_owner(self, activation):
        """Determine a user to auto-assign the task."""
        from django.contrib.auth import get_user_model

        owner = this.resolve(self.flow_class.instance, self._owner)
        if callable(owner):
            owner = owner(activation)
        elif isinstance(owner, dict):
            owner = get_user_model()._default_manager.get(**owner)
        return owner

    def calc_owner_permission(self, activation):
        """Determine required permission to assign and execute this task."""
        owner_permission = self._owner_permission
        if callable(owner_permission):
            owner_permission = owner_permission(activation)
        return owner_permission

    def calc_owner_permission_obj(self, activation):
        """Determine required permission to assign and execute this task."""
        if self._owner_permission_obj:
            return self._owner_permission_obj(activation.process)

    def can_execute(self, user, task):
        """Check user permission to execute the task."""
        return is_owner(task.owner, user)

    def can_assign(self, user, task):
        """Check if user can assign the task."""
        # already assigned
        if task.owner:
            return False

        # user not logged in
        if user is None or user.is_anonymous:
            return False

        # available for everyone
        if not task.owner_permission:
            return True

        # User have the permission
        obj = None
        if self._owner_permission_obj:
            if callable(self._owner_permission_obj):
                obj = self._owner_permission_obj(task.process)
            else:
                obj = self._owner_permission_obj

        return user.has_perm(task.owner_permission, obj=obj) or user.has_perm(
            task.owner_permission
        )

    def can_unassign(self, user, task):
        """Check if user can unassign the task."""
        # not assigned
        if task.owner is None:
            return False

        # user not logged in
        if user is None or user.is_anonymous:
            return False

        # Assigned to the same user
        if is_owner(task.owner, user):
            return True

        # User have flow management permissions
        return self.flow_class.instance.has_manage_permission(user)

    """
    TODO: Reassign
    """



import json
import traceback
import uuid

from django.utils import timezone

from ..activation import Activation
from ..base import Node
from ..fields import get_flow_ref
from ..signals import task_failed
from ..status import STATUS
from . import mixins


class AbstractJobActivation(mixins.NextNodeActivationMixin, Activation):
    @classmethod
    def create(cls, flow_task, prev_activation, token):
        """Instantiate and persist new flow task."""
        flow_class = flow_task.flow_class
        task = flow_class.task_class(
            process=prev_activation.process,
            flow_task=flow_task,
            token=token,
            external_task_id=str(uuid.uuid4()),
        )
        task.save()
        task.previous.add(prev_activation.task)
        return cls(task)

    @Activation.status.transition(source=STATUS.SCHEDULED, target=STATUS.STARTED)
    def start(self):
        self.task.started = timezone.now()
        self.task.save()

    @Activation.status.transition(source=STATUS.STARTED)
    def resume(self):
        pass

    @Activation.status.transition(source=STATUS.STARTED, target=STATUS.DONE)
    def complete(self):
        super().complete.original()

    @Activation.status.transition(source=STATUS.STARTED)
    def execute(self):
        self.complete()
        self.activate_next()

    @Activation.status.transition(source=STATUS.STARTED, target=STATUS.ERROR)
    def error(self, exception):
        if not self.task.data:
            self.task.data = {}

        tb = exception.__traceback__
        while tb.tb_next:
            tb = tb.tb_next
        serialized_locals = json.dumps(
            tb.tb_frame.f_locals, default=lambda obj: str(obj)
        )

        self.task.data["_exception"] = {
            "title": str(exception),
            "traceback": traceback.format_exc(),
            "locals": json.loads(serialized_locals),
        }
        self.task.finished = timezone.now()
        self.task.save()
        task_failed.send(sender=self.flow_class, process=self.process, task=self.task)

    def ref(self):
        return f"{get_flow_ref(self.flow_class)}/{self.process.pk}/{self.task.pk}"


class AbstractJob(mixins.NextNodeMixin, Node):
    task_type = "JOB"

    shape = {
        "width": 150,
        "height": 100,
        "text-align": "middle",
        "svg": """
            <rect class="task" width="150" height="100" rx="5" ry="5"/>
            <path class="task-label"
                  d="M19.43 12.98c.04-.32.07-.64.07-.98s-.03-.66-.07-.98l2.11-1.65
                     c.19-.15.24-.42.12-.64l-2-3.46c-.12-.22-.39-.3-.61-.22l-2.49 1
                     c-.52-.4-1.08-.73-1.69-.98l-.38-2.65 C14.46 2.18 14.25 2 14 2
                     h-4c-.25 0-.46.18-.49.42l-.38 2.65c-.61.25-1.17.59-1.69.98l-2.49-1
                     c-.23-.09-.49 0-.61.22 l-2 3.46c-.13.22-.07.49.12.64l2.11 1.65
                     c-.04.32-.07.65-.07.98s.03.66.07.98l-2.11 1.65c-.19.15-.24.42-.12.64
                     l2 3.46c.12.22.39.3.61.22l2.49-1c.52.4 1.08.73 1.69.98l.38 2.65
                     c.03.24.24.42.49.42h4c.25 0 .46-.18.49-.42l.38-2.65
                     c.61-.25 1.17-.59 1.69-.98l2.49 1c.23.09.49 0 .61-.22l2-3.46
                     c.12-.22.07-.49-.12-.64l-2.11-1.65zM12 15.5c-1.93 0-3.5-1.57-3.5-3.5
                     s1.57-3.5 3.5-3.5 3.5 1.57 3.5 3.5-1.57 3.5-3.5 3.5z"/>
        """,
    }

    bpmn_element = "scriptTask"



from typing import Callable, List, Optional, Union

from django.contrib import messages
from django.shortcuts import redirect
from django.urls import reverse
from django.utils.translation import ugettext_lazy as _


class ViewAction:
    inline_actions: Optional[List[Union[str, Callable]]] = ['view_action']

    def view_action(self, request, obj, parent_obj=None):
        """Redirect to changeform of selcted inline instance"""
        url = reverse(
            'admin:{}_{}_change'.format(
                obj._meta.app_label,
                obj._meta.model_name,
            ),
            args=(obj.pk,),
        )
        return redirect(url)

    view_action.short_description = _("View")  # type: ignore


class DeleteAction:
    def get_inline_actions(self, request, obj=None):
        actions = super().get_inline_actions(request, obj)
        if self.has_delete_permission(request, obj):
            actions.append('delete_action')
        return actions

    def delete_action(self, request, obj, parent_obj=None):
        """Remove selected inline instance if permission is sufficient"""
        if self.has_delete_permission(request):
            obj.delete()
            messages.info(request, "`{}` deleted.".format(obj))

    delete_action.short_description = _("Delete")  # type: ignore


class DefaultActionsMixin(ViewAction, DeleteAction):
    inline_actions: Optional[List[Union[str, Callable]]] = []



from django import forms
from django.contrib import messages
from django.db import transaction
from django.http import HttpResponseRedirect
from django.utils.functional import cached_property
from django.utils.translation import gettext_lazy as _
from django.views import generic

from viewflow.views import BaseBulkActionView
from viewflow.workflow.models import Task
from viewflow.workflow.status import PROCESS, STATUS
from . import mixins


class AssignTaskView(
    mixins.SuccessMessageMixin,
    mixins.TaskSuccessUrlMixin,
    mixins.TaskViewTemplateNames,
    generic.FormView,
):
    """
    Default assign view for flow task.

    Get confirmation from user, assigns task and redirects to task pages
    """

    form_class = forms.Form
    template_filename = "task_assign.html"
    success_message = _("Task {task} has been assigned.")

    def form_valid(self, *args, **kwargs):
        """If the form is valid, save the associated model and finish the task."""
        self.request.activation.assign(self.request.user)
        return super().form_valid(*args, **kwargs)


class UnassignTaskView(
    mixins.SuccessMessageMixin,
    mixins.TaskSuccessUrlMixin,
    mixins.TaskViewTemplateNames,
    generic.FormView,
):
    """
    Default unassign view for flow task.

    Get confirmation from user, and unassign task
    """

    form_class = forms.Form
    template_filename = "task_unassign.html"
    success_message = _("Task {task} has been unassigned.")

    def form_valid(self, *args, **kwargs):
        """If the form is valid, save the associated model and unassign the task."""
        self.request.activation.unassign()
        return super().form_valid(*args, **kwargs)


class CancelTaskView(
    mixins.SuccessMessageMixin,
    mixins.TaskSuccessUrlMixin,
    mixins.TaskViewTemplateNames,
    generic.FormView,
):
    """
    Default unassign view for flow task.

    Get confirmation from user, and unassign task
    """

    form_class = forms.Form
    template_filename = "task_cancel.html"
    success_message = _("Task {task} has been canceled.")

    def form_valid(self, *args, **kwargs):
        """If the form is valid, save the associated model and cancels the task."""
        self.request.activation.cancel()
        return super().form_valid(*args, **kwargs)


class UndoTaskView(
    mixins.SuccessMessageMixin,
    mixins.TaskSuccessUrlMixin,
    mixins.TaskViewTemplateNames,
    generic.FormView,
):
    """
    Default undo view for flow task.

    Get confirmation from user, and undo the task
    """

    form_class = forms.Form
    template_filename = "task_undo.html"
    success_message = _("Task {task} has been undone.")

    def form_valid(self, *args, **kwargs):
        """If the form is valid, save the associated model and undo the task."""
        self.request.activation.undo()
        return super().form_valid(*args, **kwargs)


class ReviveTaskView(
    mixins.SuccessMessageMixin,
    mixins.TaskSuccessUrlMixin,
    mixins.TaskViewTemplateNames,
    generic.FormView,
):
    """
    Default review view for flow task.

    Get confirmation from user, and revives task
    """

    form_class = forms.Form
    template_filename = "task_revive.html"
    success_message = _("Task {task} has been revived.")

    def get_success_url(self):
        if hasattr(self, "new_task"):
            return self.new_task.flow_task.reverse(
                "index", args=[self.new_task.process_id, self.new_task.pk]
            )
        return super().get_get_success_url()

    def form_valid(self, *args, **kwargs):
        """If the form is valid, save the associated model and revives the task."""
        self.new_task = self.request.activation.revive()
        return super().form_valid(*args, **kwargs)


class CancelProcessView(mixins.ProcessViewTemplateNames, generic.DetailView):
    context_object_name = "process"
    flow_class = None
    pk_url_kwarg = "process_pk"
    template_filename = "process_cancel.html"

    def get_queryset(self):
        """Flow processes."""
        return self.flow_class.process_class._default_manager.all()

    @cached_property
    def active_activations(self):
        return [
            task.flow_task.activation_class(task)
            for task in self.object.task_set.exclude(
                status__in=[STATUS.DONE, STATUS.CANCELED]
            )
        ]

    def post(self, request, *args, **kwargs):
        """Cancel active tasks and the process."""
        self.object = self.get_object()

        if self.object.status in [PROCESS.DONE, PROCESS.CANCELED]:
            messages.add_message(
                self.request,
                messages.ERROR,
                _("Process #{self.object.pk} can not be canceled."),
                fail_silently=True,
            )
            return HttpResponseRedirect("../")
        elif "_cancel_process" in request.POST:
            self.object.flow_class.instance.cancel(self.object)
            messages.add_message(
                self.request,
                messages.SUCCESS,
                _("Process #{self.object.pk} has been canceled."),
                fail_silently=True,
            )
            return HttpResponseRedirect("../")
        else:
            return self.get(request, *args, **kwargs)


class BulkUnassignTasksActionView(BaseBulkActionView):
    model = Task
    template_name = "viewflow/workflow/tasks_unassign.html"
    template_name_suffix = "s_unassign"

    def form_valid(self, form):
        with transaction.atomic():
            for task in self.get_queryset():
                with task.activation() as activation:
                    activation.unassign()
        self.message_user()
        return HttpResponseRedirect(self.get_success_url())

    def message_user(self):
        message = "Tasks was unassigned"
        messages.add_message(
            self.request, messages.SUCCESS, message, fail_silently=True
        )


class BulkAssignTasksActionView(BaseBulkActionView):
    model = Task
    template_name = "viewflow/workflow/tasks_assign.html"
    template_name_suffix = "s_assign"

    def form_valid(self, form):
        with transaction.atomic():
            for task in self.get_queryset():
                with task.activation() as activation:
                    activation.assign(self.request.user)
        self.message_user()
        return HttpResponseRedirect(self.get_success_url())

    def message_user(self):
        message = "Tasks was assigned"
        messages.add_message(
            self.request, messages.SUCCESS, message, fail_silently=True
        )



from django.apps import AppConfig


class BlackListConfig(AppConfig):
    default_auto_field = 'django.db.models.BigAutoField'
    name = 'expressmoney.apps.auth.black_list'
    verbose_name = "Черный список"



# Copyright (c) 2017-2020, Mikhail Podgurskiy
# All Rights Reserved.

# This work is dual-licensed under AGPL defined in file 'LICENSE' with
# LICENSE_EXCEPTION and the Commercial license defined in file 'COMM_LICENSE',
# which is part of this source code package.
from urllib.parse import quote as urlquote

from django.contrib import messages
from django.utils.html import format_html
from django.urls.exceptions import NoReverseMatch


class SuccessMessageMixin(object):
    """Send a notification with link to the current process or task."""

    success_message = ""

    def form_valid(self, form):
        response = super().form_valid(form)
        success_message = self.get_success_message(form.cleaned_data)
        if success_message:
            messages.success(self.request, success_message, fail_silently=True)
        return response

    def get_success_message(self, cleaned_data):
        activation = self.request.activation
        process, task = activation.process, activation.task

        try:
            process_url = process.flow_class.instance.reverse(
                "process_detail", args=[process.pk]
            )
            process_link = format_html(
                '<a href="{}">#{}</a>', urlquote(process_url), process.pk
            )
        except NoReverseMatch:
            process_link = f"#{process.pk}"

        task_url = task.flow_task.reverse("index", args=[process.pk, task.pk])
        task_link = format_html('<a href="{}">#{}</a>', urlquote(task_url), task.pk)

        return format_html(self.success_message, process=process_link, task=task_link)


class TaskSuccessUrlMixin(object):
    """Try to find .get_success_url method in the current application"""

    def get_success_url(self):
        """Continue on task or redirect back to task list."""
        if getattr(self, "success_url", None) is None:
            match = self.request.resolver_match
            if hasattr(match, "viewset") and hasattr(match.viewset, "get_success_url"):
                return match.viewset.get_success_url(self.request)

            activation = self.request.activation
            return activation.flow_task.reverse(
                "detail", args=[activation.process.pk, activation.task.pk]
            )

        return super().get_success_url()


class ProcessViewTemplateNames(object):
    """List of templates for a process view."""

    template_filename = None

    def get_template_names(self):
        if self.template_name is None:
            opts = self.flow_class.instance

            return (
                f"{opts.app_label}/{opts.flow_label}/{self.template_filename}",
                f"viewflow/workflow/{self.template_filename}",
            )
        else:
            return [self.template_name]


class TaskViewTemplateNames(object):
    """List of templates for a task view."""

    template_filename = None

    def get_template_names(self):
        if self.template_name is None:
            flow_task = self.request.activation.flow_task
            opts = flow_task.flow_class.instance
            assert not self.template_name

            return (
                f"{opts.app_label}/{opts.app_label}/{flow_task.name}_{self.template_filename}",
                f"{opts.app_label}/{opts.flow_label}/{self.template_filename}",
                f"viewflow/workflow/{self.template_filename}",
            )
        else:
            return [self.template_name]


class StoreRequestPathMixin:
    def dispatch(self, request, *args, **kwargs):
        request.session["vf-pin-location"] = request.get_full_path()
        return super().dispatch(request, *args, **kwargs)



# Copyright (c) 2017-2020, Mikhail Podgurskiy
# All Rights Reserved.

# This work is licensed under the Commercial license defined in file
# 'COMM_LICENSE', which is part of this source code package.

from enum import Enum
from typing import List, Set, Tuple
from django.db.models import Choices
from .base import State, StateDescriptor, Transition
from .typing import StateValue


def get_state_name(state_value):
    if isinstance(state_value, Enum):
        return str(state_value.value)
    return str(state_value)


def get_state_label(state_value):
    if isinstance(state_value, Choices):
        return state_value.label
    elif isinstance(state_value, Enum):
        return str(state_value.value)
    return str(state_value)


def chart(flow_state: StateDescriptor, exclude_guards=True):
    """
    Draws a directed graph (digraph) of the state transitions defined in the
    given `flow_state`.

    Args:
        flow_state (StateDescriptor): The state descriptor object.
        exclude_guards (bool): Whether to exclude transitions with no target state.

    Returns:
        str: A string representation of the digraph in the DOT language.

    The function uses the `flow_state` object to extract the transitions and the
    states involved. It then generates a DOT language string, which can be used
    with tools like Graphviz to produce an image of the graph.

    Note that the function ignores the `State.USE_RETURN_VALUE` and `State.SELECT`
    states for now, as these are not supported yet.
    """
    vertices: Set[StateValue] = set()
    postponed: List[Transition] = []
    edges: Set[Tuple[StateValue, StateValue, Transition]] = set()

    # prepare data
    for method, transitions in flow_state.get_transitions().items():
        for transition in transitions:
            if exclude_guards and transition.target is None:
                continue
            if transition.source == State.ANY:
                postponed.append(transition)
            else:
                # TODO: State.USE_RETURN_VALUE / State.SELECT
                vertices.add(transition.source)
                vertices.add(transition.target)
                edges.add((transition.source, transition.target, transition))

    for transition in postponed:
        vertices.add(transition.target)
        for vertex in vertices:
            if transition.source == State.ANY and vertex == transition.target:
                continue
            edges.add((vertex, transition.target, transition))

    # build chart
    vertices_definition = "\n".join(
        [
            '"%s" [label="%s"];' % (get_state_name(vertex), get_state_label(vertex))
            for vertex in sorted(vertices)
        ]
    )

    edges_definition = "\n".join(
        [
            f'"{source}" -> "{target}" [label="{transition.label}"];'
            for source, target, transition in sorted(edges)
        ]
    )

    return """digraph {
%s
%s
}""" % (
        vertices_definition,
        edges_definition,
    )



"""
The viewflow.fsm module provides an implementation of finite state machine (FSM)
workflows that enables users to define and execute sequential workflows with
multiple states and transitions.
"""

# Copyright (c) 2017-2020, Mikhail Podgurskiy
# All Rights Reserved.

# This work is dual-licensed under AGPL defined in file 'LICENSE' with
# LICENSE_EXCEPTION and the Commercial license defined in file 'COMM_LICENSE',
# which is part of this source code package.

from .admin import FlowAdminMixin
from .base import TransitionNotAllowed, Transition, State
from .chart import chart
from .viewset import FlowViewsMixin


__all__ = (
    "TransitionNotAllowed",
    "State",
    "FlowAdminMixin",
    "chart",
    "Transition",
    "FlowViewsMixin",
)



from django_filters import (
    DateRangeFilter,
    ChoiceFilter,
    FilterSet,
    ModelChoiceFilter as BaseModelChoiceFilter,
    MultipleChoiceFilter
)

from viewflow.this_object import this
from viewflow.workflow.status import STATUS
from viewflow.workflow.fields import get_task_ref
from viewflow.workflow.models import Process, Task


class NullDateRangeFilter(DateRangeFilter):
    def filter(self, qs, value):
        if not value:
            if not self.parent.data.get('status'):
                return qs.filter(**{f'{self.field_name}__isnull': True})
        return super().filter(qs, value)


class ModelChoiceFilter(BaseModelChoiceFilter):
    def get_queryset(self, request):
        queryset = this.resolve(self.parent, self.queryset)
        if callable(queryset):
            return queryset(request)
        return queryset


def get_queryset_flow_task_choices(queryset):
    # TODO add Node.task_name/label method
    def task_name(flow_task):
        return "{}/{}".format(flow_task.flow_class.process_title, flow_task.name.title())

    tasks = queryset.order_by('flow_task').values_list('flow_task', flat=True).distinct()
    return [
        (get_task_ref(flow_task), task_name(flow_task))
        for flow_task in tasks
    ]


class FlowUserTaskListFilter(FilterSet):
    # process = ModelChoiceFilter(queryset=this.queue_processes_query)
    flow_task = ChoiceFilter()
    created = DateRangeFilter()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.filters['flow_task'].field.choices = get_queryset_flow_task_choices(self.queryset)

    def queue_processes_query(self, request):
        return Process.objects.filter(
            pk__in=self.queryset.values('process')
        )

    class Meta:
        model = Task
        fields = ('flow_task', 'created')


class FlowArchiveListFilter(FilterSet):
    flow_task = ChoiceFilter()
    created = DateRangeFilter()
    finished = DateRangeFilter()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.filters['flow_task'].field.choices = get_queryset_flow_task_choices(self.queryset)

    class Meta:
        model = Task
        fields = ('flow_task', 'created', 'finished')


class DashboardTaskListViewFilter(FilterSet):
    flow_task = ChoiceFilter()
    status = MultipleChoiceFilter(choices=STATUS.choices)
    created = DateRangeFilter()
    finished = NullDateRangeFilter()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.filters['flow_task'].field.choices = get_queryset_flow_task_choices(self.queryset)

    class Meta:
        model = Task
        fields = ['flow_task', 'status', 'created', 'finished']


class DashboardProcessListViewFilter(FilterSet):
    created = DateRangeFilter()
    finished = NullDateRangeFilter()

    class Meta:
        model = Process
        fields = ['status', 'created', 'finished']



from django.db import models


class BankCard(models.Model):
    created = models.DateTimeField(auto_now_add=True)
    updated = models.DateTimeField(auto_now=True)
    user_id = models.PositiveIntegerField()
    is_active = models.BooleanField(default=True)
    bin = models.CharField(max_length=6)
    number = models.CharField(max_length=4, help_text='Последние 4 цифры номера карты')
    card_number = models.CharField('Номер карты', max_length=19, blank=True)
    expiry_month = models.PositiveSmallIntegerField()
    expiry_year = models.PositiveSmallIntegerField()
    ip = models.GenericIPAddressField(blank=True, null=True)
    token = models.CharField(max_length=1024, unique=True)
    comment = models.TextField(max_length=512, blank=True, help_text='Причина перевода в статус неактивной')
    # Справочная информация о карте
    brand = models.CharField('Марка', max_length=64, blank=True)
    type = models.CharField('Тип', max_length=24, blank=True)
    category = models.CharField('Категория', max_length=64, blank=True)
    issuer = models.CharField('Банк', max_length=128, blank=True)
    issuer_phone = models.CharField('Номер телефона банка', max_length=128, blank=True)
    issuer_url = models.URLField('Сайт банка', max_length=128, blank=True)
    iso_code2 = models.CharField('Страна', max_length=24, help_text='по стандарту ISO 3166-1 alpha-2', blank=True)
    iso_code3 = models.CharField('Страна', max_length=24, help_text='по стандарту ISO 3166-1 alpha-3', blank=True)
    country_name = models.CharField('Страна полное название', max_length=64, blank=True)

    def __str__(self):
        number = self.card_number
        return f'{number[:4]} {number[4:8]} {number[8:12]} {number[12:16]} {number[16:20]}'

    class Meta:
        managed = False
        verbose_name_plural = 'Банковские карты'
        verbose_name = 'Банковская карта'
        unique_together = ('bin', 'number', 'expiry_month', 'expiry_year')
        constraints = (
            models.UniqueConstraint(fields=['bin', 'number', 'expiry_month', 'expiry_year'], name='unique_bank_card'),
        )



import copy
from collections import defaultdict
from textwrap import dedent

from django.apps import apps
from django.contrib.auth import get_user_model
from django.db import transaction
from django.urls import include, path
from django.urls.exceptions import NoReverseMatch
from django.utils.timezone import now

from viewflow.urls import Viewset, ViewsetMeta
from viewflow.utils import LazySingletonDescriptor, camel_case_to_title, strip_suffixes
from . import lock
from .exceptions import FlowRuntimeError
from .status import STATUS, PROCESS


class Edge(object):
    """Edge of the Flow graph."""

    __slots__ = ("_src", "_dst", "_edge_class", "_label")

    def __init__(self, src, dst, edge_class):  # noqa D102
        self._src = src
        self._dst = dst
        self._edge_class = edge_class

    @property
    def src(self):
        """Edge source node."""
        return self._src

    @property
    def dst(self):
        """Edge destination node."""
        return self._dst

    @property
    def edge_class(self):
        """Type of the edge.

        Viewflow uses `next`, 'cond_true', `cond_false` and `default`
        edge classes.

        Edge class could be used as a hint for edge visualization.
        """
        return self._edge_class

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.src == other.src and self.dst == other.dst

    def __str__(self):
        return "[{}] {} ---> {}".format(self._edge_class, self._src, self._dst)


class Node(Viewset):
    """
    Base class for a flow task definition.

    :keyword task_type: Human readable task type
    :keyword activation_class: Activation implementation specific for this node
    """

    activation_class = None
    flow_class = None
    task_type = None

    task_title = ""
    task_description = ""
    task_summary_template = ""
    task_result_template = ""

    def __init__(self, activation_class=None, **kwargs):  # noqa D102
        self._incoming_edges = []

        if activation_class:
            self.activation_class = activation_class

        super(Node, self).__init__(**kwargs)

    @property
    def name(self):
        return self.app_name

    @name.setter
    def name(self, value):
        self.app_name = value
        if not self.task_title:
            self.task_title = value.replace("_", " ").title()

    def _get_resolver_extra(self):
        return {"node": self}

    def _outgoing(self):
        """
        Get an iterator over the outgoing edges of this node.

        :returns: An iterator over the outgoing edges of this node.
        :rtype: iterator
        """
        raise NotImplementedError

    def _incoming(self):
        """
        Get an iterator over the incoming edges of this node.

        :returns: An iterator over the incoming edges of this node.
        :rtype: iterator
        """
        return iter(self._incoming_edges)

    def __str__(self):
        if self.name:
            return self.name.title().replace("_", " ")
        return super(Node, self).__str__()

    def _resolve(self, instance):
        """Node class should resolve other nodes this-references here.

        Called as soon as node instances infatuated, but before
        incoming/outgoing links set.
        """

    def _ready(self):
        """
        Called when the flow class setup is finished.

        Subclasses can perform additional initialization here.
        """

    def _create(self, prev_activation, token):
        """
        Create a new task instance.

        :param prev_activation: The previous activation in the flow.
        :type prev_activation: Activation
        :param token: The token to use for the new task instance.
        :type token: str
        :returns: The new activation instance.
        :rtype: Activation
        """
        return self.activation_class.create(self, prev_activation, token)

    def Annotation(
        self, title=None, description=None, summary_template=None, result_template=None
    ):
        """
        Sets annotation for the node.

        :param title: The title for the task
        :param description: The description for the task
        :param summary_template: The template for the task summary
        :param result_template: The template for the task result
        :return: The node instance with the updated annotation values
        """
        if title is not None:
            self.task_title = title
        if description is not None:
            self.task_description = description
        if summary_template is not None:
            self.task_summary_template = summary_template
        if result_template is not None:
            self.task_result_template = result_template
        return self

    def _get_transition_url(self, activation, transition):
        url_name = transition.slug
        if url_name == "start":
            url_name = "execute"

        return activation.flow_task.reverse(
            url_name, args=[activation.process.pk, activation.task.pk]
        )

    def get_available_actions(self, activation, user):
        """
        Returns a list of available actions for the given user on the current node.

        :param activation: The current activation instance.
        :type activation: viewflow.models.Activation
        :param user: The user to check actions for.
        :type user: django.contrib.auth.models.User
        :return: A list of available actions as a tuple of (name, url).
        :rtype: list
        """
        # fmt: off
        transitions = activation.get_available_transitions(user)
        for transition in transitions:
            try:
                url = self._get_transition_url(activation, transition)
            except NoReverseMatch:
                pass
            else:
                yield transition.label, url


class FlowMetaClass(ViewsetMeta):
    def __str__(self):
        from .fields import get_flow_ref

        return get_flow_ref(self)


class Flow(Viewset, metaclass=FlowMetaClass):
    """
    Base class for defining a task in a flow.

    :param task_type: A human-readable string describing the task type.
    :type task_type: str
    :param activation_class: The activation class to use for this node. If not
        specified, a default activation class will be used.
    :type activation_class: class
    """

    instance = None

    process_class = None
    task_class = None
    lock_impl = lock.no_lock

    process_title = ""
    process_description = ""
    process_summary_template = ""
    process_result_template = ""

    def __init_subclass__(cls, **kwargs):
        """
        Create a new node instance.

        :param activation_class: The activation class to use for this node.
        :type activation_class: class
        :param kwargs: Additional keyword arguments to pass to the superclass.
        """
        super().__init_subclass__(**kwargs)
        cls.instance = LazySingletonDescriptor()

        # process and task default values
        from .models import Process, Task  # avoid app not loaded error

        if cls.process_class is None:
            cls.process_class = Process
        if cls.task_class is None:
            cls.task_class = Task

        # viewset.app_name
        cls.app_name = strip_suffixes(cls.__name__, ["Flow"]).lower()

        # flow description
        if not cls.process_title or not cls.process_description:
            if cls.__doc__:
                docstring = cls.__doc__.split("\n\n", 1)
                if not cls.process_title and len(docstring) > 0:
                    cls.process_title = docstring[0].strip()
                if not cls.process_description and len(docstring) > 1:
                    cls.process_description = dedent(docstring[1]).strip()
            else:
                if not cls.process_title:
                    cls.process_title = camel_case_to_title(
                        strip_suffixes(cls.__name__, ["Flow"])
                    )

        # nodes collect/copy
        cls._nodes_by_name = {}
        for attr_name in dir(cls):
            if attr_name.startswith("_") or attr_name == "instance":
                continue
            node = getattr(cls, attr_name)
            if not isinstance(node, Node):
                continue

            node = copy.copy(node)
            node.name = attr_name
            node.flow_class = cls
            cls._nodes_by_name[attr_name] = node
            setattr(cls, attr_name, node)

        # resolve inner links
        for _, node in cls._nodes_by_name.items():
            node._resolve(cls.instance)

        # setup flow graph
        incoming = defaultdict(lambda: [])  # node -> [incoming_nodes]
        for _, node in cls._nodes_by_name.items():
            for outgoing_edge in node._outgoing():
                incoming[outgoing_edge.dst].append(outgoing_edge)
        for target, edges in incoming.items():
            target._incoming_edges = edges

        # process permissions
        process_options = cls.process_class._meta
        for permission in ("manage",):
            if permission not in process_options.default_permissions:
                process_options.default_permissions += (permission,)

        # complete node setup
        for _, node in cls._nodes_by_name.items():
            node._ready()

    def __str__(self):
        return str(self.process_title)

    def _get_resolver_extra(self):
        return {"flow": self}

    @classmethod
    def lock(cls, process_pk):
        return cls.lock_impl(cls, process_pk)

    @property
    def app_label(self):
        module = "{}.{}".format(self.__module__, self.__class__.__name__)
        app_config = apps.get_containing_app_config(module)
        return app_config.label

    @property
    def flow_label(self):
        module = "{}.{}".format(self.__module__, self.__class__.__name__)
        app_config = apps.get_containing_app_config(module)

        subpath = module[len(app_config.module.__name__) + 1 :]
        if subpath.startswith("flows."):
            subpath = subpath[len("flows.") :]
        if subpath.endswith("Flow"):
            subpath = subpath[: -len("Flow")]

        return subpath.lower().replace(".", "/")

    def nodes(self):
        return self._nodes_by_name.values()

    def node(self, name, no_obsolete=False):
        node = self._nodes_by_name.get(name, None)
        if node is None and not no_obsolete:
            from .nodes import Obsolete

            obsolete_factory = self._nodes_by_name.get("obsolete", Obsolete())
            node = obsolete_factory.create_node(name, flow_class=self.__class__)
        return node

    def has_view_permission(self, user, obj=None):
        opts = self.process_class._meta
        user = get_user_model().objects.get(pk=user) if isinstance(user, int) else user
        return user.is_authenticated and user.has_perm(
            f"{opts.app_label}.view_{ opts.model_name}"
        )

    def has_manage_permission(self, user, obj=None):
        opts = self.process_class._meta
        return user.is_authenticated and user.has_perm(
            f"{opts.app_label}.manage_{ opts.model_name}"
        )

    def _get_urls(self):
        urlpatterns = super()._get_urls()

        for node in self.nodes():
            node.parent = self
            patterns, app_name, namespace = node.urls
            urlpatterns.append(
                path("", include((patterns, app_name), namespace=namespace))
            )

        return urlpatterns

    def _this_owner(self, flow_task):
        """Return user that was assigned to the task."""

        def get_task_owner(activation):
            flow_class = activation.process.flow_class
            task_class = flow_class.task_class

            task = (
                task_class._default_manager.filter(
                    process=activation.process, flow_task=flow_task, status=STATUS.DONE
                )
                .order_by("-id")
                .first()
            )
            return task.owner if task else None

        return get_task_owner

    @classmethod
    def get_start_nodes(cls, user=None):
        """
        List of flow start nodes.

        If user is not None, returns only permitted nodes for the provided user
        """
        from .nodes import Start

        return [
            node
            for node in cls.instance.nodes()
            if isinstance(node, Start)
            if user is None or node.can_execute(user)
        ]

    def get_available_process_actions(self, process, user=None):
        """List of {name, url} process actions available for the current user"""
        # TODO process cancel
        return []

    def cancel(self, process):
        with transaction.atomic(), self.lock(process.pk):
            active_tasks = process.task_set.exclude(
                status__in=[STATUS.DONE, STATUS.CANCELED]
            )

            activations = [
                task.flow_task.activation_class(task) for task in active_tasks
            ]

            not_cancellable = [
                activation
                for activation in activations
                if not activation.cancel.can_proceed()
            ]
            if not_cancellable:
                raise FlowRuntimeError(
                    "Can't cancel {}".format(
                        ",".join(activation.task for activation in not_cancellable)
                    )
                )

            for activation in activations:
                activation.cancel()

            process.status = PROCESS.CANCELED
            process.finished = now()
            process.save()



# Copyright (c) 2017-2020, Mikhail Podgurskiy
# All Rights Reserved.

# This work is dual-licensed under AGPL defined in file 'LICENSE' with
# LICENSE_EXCEPTION and the Commercial license defined in file 'COMM_LICENSE',
# which is part of this source code package.

from urllib.parse import quote as urlquote

from django.contrib import messages
from django.contrib.auth.decorators import login_required
from django.core.exceptions import PermissionDenied
from django.forms.models import modelform_factory
from django.utils.decorators import method_decorator
from django.utils.html import format_html
from django.utils.translation import gettext_lazy as _
from django.views import generic

from viewflow.forms import ModelForm, FormAjaxCompleteMixin, FormDependentSelectMixin
from viewflow.utils import viewprop, has_object_perm

from .base import FormLayoutMixin


@method_decorator(login_required, name='dispatch')
class CreateModelView(FormLayoutMixin, FormDependentSelectMixin, FormAjaxCompleteMixin, generic.CreateView):
    viewset = None
    layout = None
    form_widgets = None

    template_name_suffix = "_create"

    def has_add_permission(self, request):
        if self.viewset is not None:
            return self.viewset.has_add_permission(request.user)
        else:
            return has_object_perm(request.user, "add", self.model)

    def get_object_url(self, obj):
        if self.viewset is not None and hasattr(self.viewset, "get_object_url"):
            return self.viewset.get_object_url(self.request, obj)
        elif hasattr(obj, "get_absolute_url"):
            if has_object_perm(self.request.user, "change", obj):
                return obj.get_absolute_url()

    def message_user(self):
        url = self.get_object_url(self.object)
        link = ""
        if url:
            link = format_html('<a href="{}">{}</a>', urlquote(url), _("View"))

        message = format_html(
            _("The {obj} was added successfully. {link}"),
            obj=str(self.object),
            link=link,
        )
        messages.add_message(
            self.request, messages.SUCCESS, message, fail_silently=True
        )

    @viewprop
    def queryset(self):
        if self.viewset is not None and hasattr(self.viewset, "get_queryset"):
            return self.viewset.get_queryset(self.request)
        return None

    def get_form_class(self):
        if self.form_class is None:
            return modelform_factory(
                self.model,
                form=ModelForm,
                fields=self.fields,
                widgets=self.form_widgets,
            )
        return self.form_class

    def get_template_names(self):
        """
        List of templates for the view.
        If no `self.template_name` defined, uses::
             [<app_label>/<model_label>_<suffix>.html,
              <app_label>/<model_label>_form.html,
              'viewflow/views/form.html']
        """
        if self.template_name is None:
            opts = self.model._meta
            return [
                "{}/{}{}.html".format(
                    opts.app_label, opts.model_name, self.template_name_suffix
                ),
                "{}/{}_form.html".format(opts.app_label, opts.model_name),
                "viewflow/views/form.html",
            ]
        return [self.template_name]

    def form_valid(self, *args, **kwargs):
        response = super(CreateModelView, self).form_valid(*args, **kwargs)
        self.message_user()
        return response

    def get_success_url(self):
        if self.viewset and hasattr(self.viewset, "get_success_url"):
            return self.viewset.get_success_url(self.request, obj=self.object)
        return "../"

    def dispatch(self, request, *args, **kwargs):
        if not self.has_add_permission(self.request):
            raise PermissionDenied

        return super(CreateModelView, self).dispatch(request, *args, **kwargs)



"""Prevents inconsistent db updates for flow."""
from __future__ import unicode_literals

import time
import random
from contextlib import contextmanager

from django.core.cache import cache as default_cache
from django.db import transaction, DatabaseError

from .exceptions import FlowLockFailed


class NoLock(object):
    """
    No pessimistic locking, just execute flow task in transaction.

    Not suitable when you have Join nodes in your flow.
    """

    @contextmanager
    def __call__(self, flow_class, process_pk):
        with transaction.atomic():
            yield


class SelectForUpdateLock(object):
    """
    Database lock uses `select ... for update` on the process instance row.

    Recommended to use with PostgreSQL.
    """
    def __init__(self, nowait=True, attempts=5):
        self.nowait = nowait
        self.attempts = attempts

    @contextmanager
    def __call__(self, flow_class, process_pk):
        for i in range(self.attempts):
            with transaction.atomic():
                try:
                    process = flow_class.process_class._default_manager.filter(pk=process_pk)
                    process.select_for_update(nowait=self.nowait).exists()
                except DatabaseError:
                    if i != self.attempts - 1:
                        sleep_time = (((i + 1) * random.random()) + 2 ** i) / 2.5
                        time.sleep(sleep_time)
                    else:
                        raise FlowLockFailed('Lock failed for {}'.format(flow_class))
                else:
                    yield
                    break


class CacheLock(object):
    """
    Task lock based on Django cache.

    Use it if primary cache backend has transactional `add` functionality,
    like `memcached`.

    Example::

        class MyFlow(Flow):
            lock_impl = CacheLock(cache=caches['locks'])

    The example uses a different cache. The default cache
    is Django ``default`` cache configuration.
    """

    def __init__(self, cache=default_cache, attempts=5, expires=120):  # noqa D102
        self.cache = cache
        self.attempts = attempts
        self.expires = expires

    @contextmanager
    def __call__(self, flow_class, process_pk):  # noqa D102
        key = 'django-viewflow-lock-{}/{}'.format(flow_class.instance.flow_label, process_pk)

        for i in range(self.attempts):
            if self.cache.add(key, 1, self.expires):
                break
            if i != self.attempts - 1:
                sleep_time = (((i + 1) * random.random()) + 2 ** i) / 2.5
                time.sleep(sleep_time)
        else:
            raise FlowLockFailed('Lock failed for {}'.format(flow_class))

        try:
            with transaction.atomic():
                yield
        finally:
            self.cache.delete(key)


no_lock = NoLock()
cache_lock = CacheLock()
select_for_update_lock = SelectForUpdateLock()



"""Viewflow - The Django extension for perfectionists with yesterday’s deadlines."""

# Copyright (c) 2017-2020, Mikhail Podgurskiy
# All Rights Reserved.

# This work is dual-licensed under AGPL defined in file 'LICENSE' with
# LICENSE_EXCEPTION and the Commercial license defined in file 'COMM_LICENSE',
# which is part of this source code package.

from django.conf import settings as django_settings
from .conf import settings
from .this_object import this
from .utils import viewprop, Icon, DEFAULT

__title__ = 'Django-Viewflow'
__version__ = '2.0a2'
__author__ = 'Mikhail Podgurskiy'
__license__ = 'AGPL'
__copyright__ = 'Copyright 2018-2021 Mikhail Podgurskiy'

__all__ = (
    'this', 'viewprop', 'Icon', 'DEFAULT'
)

default_app_config = 'viewflow.apps.ViewflowConfig'

if settings.AUTOREGISTER:
    # Register site middleware
    site_middleware = 'viewflow.middleware.SiteMiddleware'
    if site_middleware not in django_settings.MIDDLEWARE:
        django_settings.MIDDLEWARE += (site_middleware, )
    turbo_middleware = 'viewflow.middleware.HotwireTurboMiddleware'
    if turbo_middleware not in django_settings.MIDDLEWARE:
        django_settings.MIDDLEWARE += (turbo_middleware, )



from django.utils.timezone import now

from viewflow import this
from ..activation import Activation
from ..base import Node
from ..status import STATUS
from ..signals import task_started
from . import mixins


class FunctionActivation(mixins.NextNodeActivationMixin, Activation):
    """
    Handle  Activation.

    Executes a callback immediately.
    """

    @Activation.status.super()
    def activate(self):
        """Perform the callback within current exception propagation strategy."""
        with self.exception_guard():
            self.task.started = now()
            task_started.send(
                sender=self.flow_class, process=self.process, task=self.task
            )
            self.flow_task._func(self)

    @Activation.status.transition(source=STATUS.ERROR, target=STATUS.DONE)
    def retry(self):
        """Retry the next node calculation and activation."""
        self.activate.original()


class Function(mixins.NextNodeMixin, mixins.NodePermissionMixin, Node):
    """
    Callback executed synchronously on a task activation.
    """

    activation_class = FunctionActivation

    task_type = "FUNCTION"

    shape = {
        "width": 150,
        "height": 100,
        "text-align": "middle",
        "svg": """
            <rect class="task" width="150" height="100" rx="5" ry="5"/>
            <path class="task-marker"
                  d="m 15,20 c 10,-6 -8,-8 3,-15 l -9,0 c -11,7 7,9 -3,15 z
                     m -7,-12 l 5,0 m -4.5,3 l 4.5,0 m -3,3 l 5,0 m -4,3 l 5,0"/>
        """,
    }

    bpmn_element = "scriptTask"

    def __init__(self, func, **kwargs):  # noqa D102
        super().__init__(**kwargs)
        self._func = func

    def _resolve(self, instance):
        super()._resolve(instance)
        self._func = this.resolve(instance, self._func)



from django.db import models


class NBKICreditHistoryV2(models.Model):
    """
    from cache.models import NBKICreditHistoryV2
    """
    created = models.DateTimeField(auto_now_add=True)
    xml = models.URLField(blank=True, help_text='Ссылка на текущую КИ в формате xml')
    time = models.DecimalField(max_digits=9, decimal_places=4, null=True, help_text='Время загрузки КИ')
    # Personal data
    first_name = models.CharField(max_length=32)
    last_name = models.CharField(max_length=32)
    middle_name = models.CharField(max_length=32)
    birth_date = models.DateField()
    passport_serial = models.CharField(max_length=4)
    passport_number = models.CharField(max_length=6)
    passport_date = models.DateField()
    passport_code = models.CharField(max_length=7, blank=True)
    snils = models.CharField(max_length=64, blank=True)
    # NBKI features
    accounts_total_mfo_3y = models.IntegerField(null=True, help_text='Всего счетов МФО за 3 года')
    accounts_total_mfo_gt3y = models.IntegerField(null=True, help_text='Всего счетов МФО от 3 лет')
    accounts_total_bank_3y = models.IntegerField(null=True, help_text='Всего счетов БАНК за 3 года')
    accounts_total_bank_gt3y = models.IntegerField(null=True, help_text='Всего счетов БАНК от 3 лет')
    accounts_active_mfo = models.IntegerField(null=True, help_text='Действующих счетов МФО')
    accounts_active_bank = models.IntegerField(null=True, help_text='Действующих счетов БАНК')
    accounts_zero = models.IntegerField(null=True, help_text='Действующих счетов без платежей')
    accounts_pastdue_mfo = models.IntegerField(null=True, help_text='Просрочено счетов МФО')
    accounts_pastdue_bank = models.IntegerField(null=True, help_text='Просрочено счетов БАНК')
    accounts_closed_mfo_3y = models.IntegerField(null=True, help_text='Всего закрыто МФО за 3 года')
    accounts_closed_mfo_gt3y = models.IntegerField(null=True, help_text='Всего закрыто МФО от 3 лет')
    accounts_closed_bank_3y = models.IntegerField(null=True, help_text='Всего закрыто БАНК за 3 года')
    accounts_closed_bank_gt3y = models.IntegerField(null=True, help_text='Всего закрыто БАНК от 3 лет')
    balance_credit_mfo_3y = models.IntegerField(null=True, help_text='Выдано тело МФО за 3 года')
    balance_credit_mfo_gt3y = models.IntegerField(null=True, help_text='Выдано тело МФО от 3 лет')
    balance_credit_bank_3y = models.IntegerField(null=True, help_text='Выдано тело БАНК за 3 года')
    balance_credit_bank_gt3y = models.IntegerField(null=True, help_text='Выдано тело БАНК от 3 лет')
    balance_debt_mfo = models.IntegerField(null=True, help_text='Текущий долг по займам МФО')
    balance_debt_bank = models.IntegerField(null=True, help_text='Текущий долг по кредитам БАНК')
    balance_nextpay_mfo = models.IntegerField(null=True, help_text='Следующий платеж по займам МФО')
    balance_nextpay_bank = models.IntegerField(null=True, help_text='Следующий платеж по кредитам БАНК')
    balance_paid_mfo_3y = models.IntegerField(null=True, help_text='Выплачено по займам МФО за 3 года')
    balance_paid_mfo_gt3y = models.IntegerField(null=True, help_text='Выплачено по займам МФО от 3 лет')
    balance_paid_bank_3y = models.IntegerField(null=True, help_text='Выплачено по кредитам БАНК за 3 года')
    balance_paid_bank_gt3y = models.IntegerField(null=True, help_text='Выплачено по кредитам БАНК от 3 лет')
    balance_pastdue_mfo = models.IntegerField(null=True, help_text='Текущая просрочка по займам МФО')
    balance_pastdue_bank = models.IntegerField(null=True, help_text='Текущая просрочка по кредитам БАНК')
    inquiries_total = models.IntegerField(null=True, help_text='Всего запросов на получение займов и кредитов')
    inquiries_amount_mfo = models.IntegerField(null=True, help_text='Средняя запрашиваемая сумма в МФО')
    inquiries_amount_bank = models.IntegerField(null=True, help_text='Средняя запрашиваемая сумма в Банке')
    inquiries_7 = models.IntegerField(null=True, help_text='Запросы за последние 7 дней')
    inquiries_14 = models.IntegerField(null=True, help_text='Запросы за последние 14 дней')
    inquiries_21 = models.IntegerField(null=True, help_text='Запросы за последние 21 дней')
    inquiries_30 = models.IntegerField(null=True, help_text='Запросы за последние 30 дней')
    inquiries_60 = models.IntegerField(null=True, help_text='Запросы за последние 60 дней')
    inquiries_90 = models.IntegerField(null=True, help_text='Запросы за последние 90 дней')
    inquiries_180 = models.IntegerField(null=True, help_text='Запросы за последние 180 дней')
    inquiries_365 = models.IntegerField(null=True, help_text='Запросы за последние 365 дней')
    rate_inquiries = models.DecimalField(max_digits=3, decimal_places=2, null=True, help_text='Счета/Запросы')
    rate_inquiries_7 = models.DecimalField(max_digits=3, decimal_places=2, null=True,
                                           help_text='Всего счетов за 3г к запросам за 7 дней')
    rate_inquiries_7_365 = models.DecimalField(max_digits=3, decimal_places=2, null=True,
                                               help_text='Всего запросов к запросам за 7 дней')
    rate_inquiries_14 = models.DecimalField(max_digits=3, decimal_places=2, null=True,
                                            help_text='Всего счетов за 3г к запросам за 14 дней')
    rate_inquiries_14_365 = models.DecimalField(max_digits=3, decimal_places=2, null=True,
                                                help_text='Всего запросов к запросам за 14 дней')
    rate_closed = models.DecimalField(max_digits=3, decimal_places=2, null=True, help_text='Закрыто сч./Всего счетов')
    rate_paid = models.DecimalField(max_digits=3, decimal_places=2, null=True, help_text='Выплачено / Всего выдано')
    rate_debt = models.DecimalField(max_digits=3, decimal_places=2, null=True, help_text='Долг/выдано')
    rate_pastdue = models.DecimalField(max_digits=3, decimal_places=2, null=True, help_text='Просрочено/Выдано')
    rate_pastdue_debt = models.DecimalField(max_digits=3, decimal_places=2, null=True,
                                            help_text='Просроченная задолженность к всего задолженность')
    rate_zero = models.DecimalField(max_digits=3, decimal_places=2, null=True,
                                    help_text='Счето без оплат к всего счетов за 3 года')
    rate_zero_active = models.DecimalField(max_digits=3, decimal_places=2, null=True,
                                           help_text='Счето без оплат к активным')
    rate_active = models.DecimalField(max_digits=3, decimal_places=2, null=True,
                                      help_text='Активных к всего счетов за 3 года')
    rate_accounts_pastdue = models.DecimalField(max_digits=3, decimal_places=2, null=True,
                                                help_text='Просроченных к всего счетам за 3 года')
    rate_accounts_pastdue_active = models.DecimalField(max_digits=3, decimal_places=2, null=True,
                                                       help_text='Просроченных к активным счетам за 3 года')
    avg_credit_mfo_3y = models.IntegerField(null=True, help_text='Средняя сумма займа за последние 3 года')
    avg_credit_bank_3y = models.IntegerField(null=True, help_text='Средняя сумма кредита за последние 3 года')
    max_credit_amount_mfo_3y = models.IntegerField(null=True, help_text='Максимальная сумма займа/кредита')
    max_credit_amount_mfo_gt3y = models.IntegerField(null=True, help_text='Максимальная сумма займа/кредита')
    max_credit_amount_bank_3y = models.IntegerField(null=True, help_text='Максимальная сумма займа/кредита')
    max_credit_amount_bank_gt3y = models.IntegerField(null=True, help_text='Максимальная сумма займа/кредита')
    interests_mfo_3y = models.IntegerField(null=True, help_text='Выплачено процентов по закрытым счетам')
    interests_mfo_gt3y = models.IntegerField(null=True, help_text='Выплачено процентов по закрытым счетам')
    interests_bank_3y = models.IntegerField(null=True, help_text='Выплачено процентов по закрытым счетам')
    interests_bank_gt3y = models.IntegerField(null=True, help_text='Выплачено процентов по закрытым счетам')
    interests_avg_mfo_3y = models.IntegerField(null=True, help_text='В среднем выплачено процентов на 1 займ/кредит')
    interests_avg_mfo_gt3y = models.IntegerField(null=True, help_text='В среднем выплачено процентов на 1 займ/кредит')
    interests_avg_bank_3y = models.IntegerField(null=True, help_text='В среднем выплачено процентов на 1 займ/кредит')
    interests_avg_bank_gt3y = models.IntegerField(null=True, help_text='В среднем выплачено процентов на 1 займ/кредит')
    days_last = models.IntegerField(null=True, help_text='Дней с последнего займа/кредита')
    days_between = models.IntegerField(null=True, help_text='Дней между займами/кредитами')

    def save(self, *args, **kwargs):
        pass

    def __str__(self):
        return str(self.pk)

    class Meta:
        managed = False
        verbose_name_plural = 'NBKI credit history v2'
        ordering = ('-created',)


class NBKIScore(models.Model):
    created = models.DateTimeField(auto_now_add=True)
    score = models.PositiveSmallIntegerField()
    # Passport
    first_name = models.CharField(max_length=32)
    last_name = models.CharField(max_length=32)
    middle_name = models.CharField(max_length=32)
    birth_date = models.DateField()
    passport_serial = models.CharField(max_length=4)
    passport_number = models.CharField(max_length=6)
    passport_date = models.DateField()
    passport_code = models.CharField(max_length=7, blank=True)
    snils = models.CharField(max_length=64, blank=True)

    def save(self, *args, **kwargs):
        pass

    def __str__(self):
        return str(self.pk)

    class Meta:
        managed = False



from django.views.decorators.http import last_modified
from django.http import HttpResponse
from django.utils import timezone
from django.utils.decorators import method_decorator
from django.views import generic
from ... import chart


process_start_time = timezone.now()


class FlowChartView(generic.View):
    """Flow and Process chart View."""

    flow_class = None

    @method_decorator(
        last_modified(
            lambda *args, **kwargs: None
            if "process_pk" in kwargs
            else process_start_time
        )
    )
    def get(self, request, *args, **kwargs):
        process_pk = kwargs.get("process_pk")

        grid = chart.calc_layout_data(self.flow_class)
        if process_pk is not None:
            chart.calc_cell_status(self.flow_class, grid, process_pk)
        svg = chart.grid_to_svg(grid)

        return HttpResponse(svg, content_type="image/svg+xml")



from django.apps import AppConfig


class OperationsConfig(AppConfig):
    default_auto_field = 'django.db.models.BigAutoField'
    label = 'ru_loans_operations'
    name = 'expressmoney.apps.ru_loans.operations'
    verbose_name = 'Операции по займам'



from django.db import models


class BlackList(models.Model):
    EMPTY = 'EMPTY'
    SCORING = 'SCORING'
    FRAUD = 'FRAUD'
    NOT_INTEREST = 'NOT_INTEREST'
    LEGAL = 'LEGAL'
    SPAM = 'SPAM'

    CAUSE_CHOICES = (
        (EMPTY, 'Не указана'),
        (SCORING, 'Низкий скор'),
        (FRAUD, 'Мошенник'),
        (NOT_INTEREST, 'Просил заблокировать'),
        (LEGAL, 'Юр. ограничения'),
        (SPAM, 'Бот/Спам'),

    )
    created = models.DateTimeField('Добавлен', auto_now_add=True)
    user_id = models.CharField(blank=True, max_length=128)
    cause = models.CharField('Причина', max_length=16, choices=CAUSE_CHOICES, default=EMPTY, null=True)
    passport_serial = models.CharField('Серия паспорта', max_length=4, blank=True)
    passport_number = models.CharField('Номер паспорта', max_length=6, blank=True)
    comment = models.CharField('Комментарий', max_length=128, blank=True)

    def save(self, *args, **kwargs):
        if self._state.adding and self.user_id and WhiteList.objects.filter(pk=self.user_id).exists():
            pass
        else:
            super().save(*args, **kwargs)

    def __str__(self):
        return f'{self.user_id}'

    class Meta:
        managed = False
        verbose_name = 'Черный список'
        verbose_name_plural = 'Черный список'
        constraints = (
            models.UniqueConstraint(fields=('passport_serial', 'passport_number'),
                                    condition=(~models.Q(passport_serial__exact='') &
                                               ~models.Q(passport_number__exact='')),
                                    name='unique_passport_black_list'),
            models.UniqueConstraint(fields=('user_id',), condition=(~models.Q(user_id__exact='')), name='unique_user')
        )


class WhiteList(models.Model):
    created = models.DateTimeField('Добавлен', auto_now_add=True)
    user_id = models.PositiveIntegerField(primary_key=True)
    comment = models.CharField('Комментарий', max_length=128, blank=True)

    def __str__(self):
        return f'{self.user_id}'

    class Meta:
        managed = False
        verbose_name = 'Белый список'
        verbose_name_plural = 'Белый список'



from django.core.validators import MinLengthValidator, MaxValueValidator, MinValueValidator
from django.db import models

from expressmoney.apps.ru_scoring.cache.models import NBKIScore, NBKICreditHistoryV2


class Scoring(models.Model):
    """
    from pdl_ru.models import Scoring
    obj = Scoring.objects.create(**{
        "amount_requested": 16_000,
        "period_requested": 30,
        "first_name": "Султан",
        "last_name": "Шамурзаев",
        "middle_name": "Хазретович",
        "birth_date": "1999-08-22",
        "passport_serial": "9019",
        "passport_number": "314603",
        "passport_code": "150-010",
        "passport_date": "2019-09-25",
        "snils": "15460066042",
        "total_loans": 0,
        "interests_sum": 0,
        "body_sum": 0,
        "body_max": 0,
        "department": "zaem.ru"
    })
    """

    created = models.DateTimeField(auto_now_add=True)
    # Order
    order_id = models.CharField(max_length=32, unique=True, validators=(MinLengthValidator(1),), null=True)
    user_id = models.PositiveIntegerField(null=True)
    period_requested = models.PositiveSmallIntegerField(validators=(MinValueValidator(1), MaxValueValidator(500)))
    amount_requested = models.PositiveIntegerField(validators=(MinValueValidator(1000), MaxValueValidator(200000)))
    # Passport
    first_name = models.CharField(max_length=32, validators=(MinLengthValidator(1),))
    last_name = models.CharField(max_length=32, validators=(MinLengthValidator(1),))
    middle_name = models.CharField(max_length=32)
    birth_date = models.DateField()
    passport_serial = models.CharField(max_length=4, validators=(MinLengthValidator(4),))
    passport_number = models.CharField(max_length=6, validators=(MinLengthValidator(6),))
    passport_date = models.DateField()
    passport_code = models.CharField(max_length=7, validators=(MinLengthValidator(7),))
    snils = models.CharField(max_length=16)
    # Loans data
    total_loans = models.PositiveIntegerField(help_text='Всего займов')
    interests_sum = models.PositiveIntegerField(help_text='Всего выплачено процентов')
    body_sum = models.PositiveIntegerField(help_text='Всего выплачено тело')
    body_max = models.PositiveIntegerField(help_text='Максимально выданная ранее сумма')
    first_loan = models.IntegerField(null=True, help_text='Дней с первого займа')
    last_loan = models.IntegerField(null=True, help_text='Дней с последнего займа')
    between_loans = models.IntegerField(null=True, help_text='Ср. кол. дней между займами')
    # Data
    score = models.DecimalField(max_digits=3, decimal_places=2, null=True)
    score_min = models.DecimalField(max_digits=3, decimal_places=2, help_text='Мин. скор для одобр. суммы', null=True)
    amount_approved = models.PositiveIntegerField(null=True)
    department = models.CharField(max_length=128, default='', help_text='Бизнес-юнит. Например: займы.рф', null=True)
    cache_ki = models.ForeignKey(NBKICreditHistoryV2,
                                 models.SET_NULL,
                                 null=True,
                                 blank=True,
                                 related_name='%(app_label)s_%(class)s')
    cache_score = models.ForeignKey(NBKIScore,
                                    models.SET_NULL,
                                    null=True,
                                    blank=True,
                                    related_name='%(app_label)s_%(class)s')
    nbki_score = models.PositiveIntegerField(null=True)
    xml = models.CharField(max_length=512, blank=True)
    time = models.DecimalField(max_digits=9, decimal_places=4, null=True, help_text='Время расчета скора, секунды')
    is_black_list = models.BooleanField(null=True)
    is_offline = models.BooleanField(null=True, help_text='Минимум один раз лично посещал офис')
    comment = models.TextField(max_length=1024, blank=True)

    def save(self, *args, **kwargs):
        pass

    class Meta:
        managed = False


class FeatureStore(models.Model):
    order_id = models.CharField(primary_key=True, max_length=32)
    is_from_scoring = models.BooleanField(default=True, help_text='False - это займы примерно до конца мая 2023, '
                                                                  'их нет в Scoring. (71 025  шт.) '
                                                                  'Удалять нельзя так как не восстановить.')
    is_refilled = models.BooleanField(default=True, help_text='True - фичи перезаполнены. Сбросить перед заполнением.')
    revenue = models.DecimalField(max_digits=10, decimal_places=2, null=True,
                                  help_text='Сумма выплаченных процентов с момента выдачи займа на момент заполнения')
    ENDPOINT_ID = None
    MODEL_NAME = None
    explanation = None
    last_ki = None
    _max_inquiries_7 = 15
    # Order
    order_created = models.DateTimeField()
    period_requested = models.PositiveSmallIntegerField()
    amount_requested = models.PositiveIntegerField()
    amount_approved = models.PositiveIntegerField()
    # Other
    comment = models.CharField(max_length=1024, default='')
    time = models.DecimalField(max_digits=9, decimal_places=4, null=True, help_text='Время расчета скора, секунды')
    department = models.CharField(max_length=32, blank=True)
    is_offline = models.BooleanField(null=True, help_text='Минимум один раз лично посещал офис')
    # User
    user_id = models.PositiveIntegerField(null=True)
    # Profile
    first_name = models.CharField(max_length=32, default='')
    last_name = models.CharField(max_length=32, default='')
    middle_name = models.CharField(max_length=32, default='')
    birth_date = models.DateField(null=True)
    passport_serial = models.CharField(max_length=4, default='')
    passport_number = models.CharField(max_length=6, default='')
    passport_date = models.DateField(null=True)
    passport_code = models.CharField(max_length=7, blank=True)
    snils = models.CharField(max_length=16, default='')
    # Loan
    loan_id = models.PositiveIntegerField(null=True, unique=True)
    loan_created = models.DateField(null=True)
    loan_status = models.CharField(max_length=16, null=True)
    loan_status_date = models.DateField(null=True)
    # Expressmoney features
    em_total_loans = models.IntegerField(null=True, help_text='Всего займов')
    em_between_loans = models.IntegerField(null=True, help_text='Ср. кол. дней между займами')
    em_first_loan = models.IntegerField(null=True, help_text='Дней с первого займа')
    em_last_loan = models.IntegerField(null=True, help_text='Дней с последнего займа')
    em_interests_sum = models.IntegerField(null=True, help_text='Выплачено процентов по всем предыдущим займам')
    em_interests_mean = models.IntegerField(null=True, help_text='В среднем выплачено процентов на 1 займ')
    em_interests_rate = models.IntegerField(null=True, help_text='Выплачено процентов / Одобренная сумма')
    em_body_sum = models.IntegerField(null=True, help_text='Сумма тела всех займов')
    em_body_mean = models.IntegerField(null=True, help_text='Средняя сумма займа')
    em_body_mean_rate = models.IntegerField(null=True, help_text='Средняя сумма тела / Одобренная сумма')
    em_body_max = models.IntegerField(null=True, help_text='Максимальная сумма займа')
    em_body_max_rate = models.IntegerField(null=True, help_text='Максимальная сумма займа / Одобренная сумма')
    # NBKI score
    nbki_score = models.IntegerField(null=True, help_text='Больше - лучше')
    # NBKI features
    accounts_30d = models.IntegerField(null=True, help_text='Всего счетов МФО за 30 дней')
    accounts_mfo_30d = models.IntegerField(null=True, help_text='Всего счетов МФО за 30 дней')
    accounts_mfo_30d_rate = models.DecimalField(max_digits=3, decimal_places=2, null=True, help_text='Доля МФО 30 дней')
    accounts_total_mfo_3y = models.IntegerField(null=True, help_text='Всего счетов МФО за 3 года')
    accounts_total_mfo_gt3y = models.IntegerField(null=True, help_text='Всего счетов МФО от 3 лет')
    accounts_total_bank_3y = models.IntegerField(null=True, help_text='Всего счетов БАНК за 3 года')
    accounts_total_bank_gt3y = models.IntegerField(null=True, help_text='Всего счетов БАНК от 3 лет')
    accounts_active_mfo = models.IntegerField(null=True, help_text='Действующих счетов МФО')
    accounts_active_bank = models.IntegerField(null=True, help_text='Действующих счетов БАНК')
    accounts_zero = models.IntegerField(null=True, help_text='Действующих счетов без платежей')
    accounts_pastdue_mfo = models.IntegerField(null=True, help_text='Просрочено счетов МФО')
    accounts_pastdue_bank = models.IntegerField(null=True, help_text='Просрочено счетов БАНК')
    accounts_closed_mfo_3y = models.IntegerField(null=True, help_text='Всего закрыто МФО за 3 года')
    accounts_closed_mfo_gt3y = models.IntegerField(null=True, help_text='Всего закрыто МФО от 3 лет')
    accounts_closed_bank_3y = models.IntegerField(null=True, help_text='Всего закрыто БАНК за 3 года')
    accounts_closed_bank_gt3y = models.IntegerField(null=True, help_text='Всего закрыто БАНК от 3 лет')
    balance_credit_mfo_3y = models.IntegerField(null=True, help_text='Выдано тело МФО за 3 года')
    balance_credit_mfo_gt3y = models.IntegerField(null=True, help_text='Выдано тело МФО от 3 лет')
    balance_credit_bank_3y = models.IntegerField(null=True, help_text='Выдано тело БАНК за 3 года')
    balance_credit_bank_gt3y = models.IntegerField(null=True, help_text='Выдано тело БАНК от 3 лет')
    balance_debt_mfo = models.IntegerField(null=True, help_text='Текущий долг по займам МФО')
    balance_debt_bank = models.IntegerField(null=True, help_text='Текущий долг по кредитам БАНК')
    balance_nextpay_mfo = models.IntegerField(null=True, help_text='Следующий платеж по займам МФО')
    balance_nextpay_bank = models.IntegerField(null=True, help_text='Следующий платеж по кредитам БАНК')
    balance_paid_mfo_3y = models.IntegerField(null=True, help_text='Выплачено по займам МФО за 3 года')
    balance_paid_mfo_gt3y = models.IntegerField(null=True, help_text='Выплачено по займам МФО от 3 лет')
    balance_paid_bank_3y = models.IntegerField(null=True, help_text='Выплачено по кредитам БАНК за 3 года')
    balance_paid_bank_gt3y = models.IntegerField(null=True, help_text='Выплачено по кредитам БАНК от 3 лет')
    balance_pastdue_mfo = models.IntegerField(null=True, help_text='Текущая просрочка по займам МФО')
    balance_pastdue_bank = models.IntegerField(null=True, help_text='Текущая просрочка по кредитам БАНК')
    inquiries_total = models.IntegerField(null=True, help_text='Всего запросов на получение займов и кредитов')
    inquiries_amount_mfo = models.IntegerField(null=True, help_text='Средняя запрашиваемая сумма в МФО')
    inquiries_amount_bank = models.IntegerField(null=True, help_text='Средняя запрашиваемая сумма в Банке')
    inquiries_7 = models.IntegerField(null=True, help_text='Запросы за последние 7 дней')
    inquiries_14 = models.IntegerField(null=True, help_text='Запросы за последние 14 дней')
    inquiries_21 = models.IntegerField(null=True, help_text='Запросы за последние 21 дней')
    inquiries_30 = models.IntegerField(null=True, help_text='Запросы за последние 30 дней')
    inquiries_60 = models.IntegerField(null=True, help_text='Запросы за последние 60 дней')
    inquiries_90 = models.IntegerField(null=True, help_text='Запросы за последние 90 дней')
    inquiries_180 = models.IntegerField(null=True, help_text='Запросы за последние 180 дней')
    inquiries_365 = models.IntegerField(null=True, help_text='Запросы за последние 365 дней')
    rate_inquiries = models.DecimalField(max_digits=3, decimal_places=2, null=True, help_text='Счета/Запросы')
    rate_inquiries_7 = models.DecimalField(max_digits=3, decimal_places=2, null=True,
                                           help_text='Всего счетов за 3г к запросам за 7 дней')
    rate_inquiries_7_365 = models.DecimalField(max_digits=3, decimal_places=2, null=True,
                                               help_text='Всего запросов к запросам за 7 дней')
    rate_inquiries_14 = models.DecimalField(max_digits=3, decimal_places=2, null=True,
                                            help_text='Всего счетов за 3г к запросам за 14 дней')
    rate_inquiries_14_365 = models.DecimalField(max_digits=3, decimal_places=2, null=True,
                                                help_text='Всего запросов к запросам за 14 дней')
    rate_closed = models.DecimalField(max_digits=3, decimal_places=2, null=True, help_text='Закрыто сч./Всего счетов')
    rate_paid = models.DecimalField(max_digits=3, decimal_places=2, null=True, help_text='Выплачено / Всего выдано')
    rate_debt = models.DecimalField(max_digits=3, decimal_places=2, null=True, help_text='Долг/выдано')
    rate_pastdue = models.DecimalField(max_digits=3, decimal_places=2, null=True, help_text='Просрочено/Выдано')
    rate_pastdue_debt = models.DecimalField(max_digits=3, decimal_places=2, null=True,
                                            help_text='Просроченная задолженность к всего задолженность')
    rate_zero = models.DecimalField(max_digits=3, decimal_places=2, null=True,
                                    help_text='Счето без оплат к всего счетов за 3 года')
    rate_zero_active = models.DecimalField(max_digits=3, decimal_places=2, null=True,
                                           help_text='Счето без оплат к активным')
    rate_active = models.DecimalField(max_digits=3, decimal_places=2, null=True,
                                      help_text='Активных к всего счетов за 3 года')
    rate_accounts_pastdue = models.DecimalField(max_digits=3, decimal_places=2, null=True,
                                                help_text='Просроченных к всего счетам за 3 года')
    rate_accounts_pastdue_active = models.DecimalField(max_digits=3, decimal_places=2, null=True,
                                                       help_text='Просроченных к активным счетам за 3 года')
    avg_credit_mfo_3y = models.IntegerField(null=True, help_text='Средняя сумма займа за последние 3 года')
    avg_credit_bank_3y = models.IntegerField(null=True, help_text='Средняя сумма кредита за последние 3 года')
    max_credit_amount_mfo_3y = models.IntegerField(null=True, help_text='Максимальная сумма займа/кредита')
    max_credit_amount_mfo_gt3y = models.IntegerField(null=True, help_text='Максимальная сумма займа/кредита')
    max_credit_amount_bank_3y = models.IntegerField(null=True, help_text='Максимальная сумма займа/кредита')
    max_credit_amount_bank_gt3y = models.IntegerField(null=True, help_text='Максимальная сумма займа/кредита')
    interests_mfo_3y = models.IntegerField(null=True, help_text='Выплачено процентов по закрытым счетам')
    interests_mfo_gt3y = models.IntegerField(null=True, help_text='Выплачено процентов по закрытым счетам')
    interests_bank_3y = models.IntegerField(null=True, help_text='Выплачено процентов по закрытым счетам')
    interests_bank_gt3y = models.IntegerField(null=True, help_text='Выплачено процентов по закрытым счетам')
    interests_avg_mfo_3y = models.IntegerField(null=True, help_text='В среднем выплачено процентов на 1 займ/кредит')
    interests_avg_mfo_gt3y = models.IntegerField(null=True, help_text='В среднем выплачено процентов на 1 займ/кредит')
    interests_avg_bank_3y = models.IntegerField(null=True, help_text='В среднем выплачено процентов на 1 займ/кредит')
    interests_avg_bank_gt3y = models.IntegerField(null=True, help_text='В среднем выплачено процентов на 1 займ/кредит')
    days_last = models.IntegerField(null=True, help_text='Дней с последнего займа/кредита')
    days_between = models.IntegerField(null=True, help_text='Дней между займами/кредитами')

    # Source
    xml = models.CharField(max_length=512, default='', help_text='Название xml использованного для заполнения фичей')
    cache_ki = models.ForeignKey(NBKICreditHistoryV2,
                                 models.SET_NULL,
                                 null=True,
                                 blank=True,
                                 related_name='%(app_label)s_%(class)s')
    cache_score = models.ForeignKey(NBKIScore,
                                    models.SET_NULL,
                                    null=True,
                                    blank=True,
                                    related_name='%(app_label)s_%(class)s')

    class Meta:
        managed = False


# Copyright (c) 2017-2020, Mikhail Podgurskiy
# All Rights Reserved.

# This work is dual-licensed under AGPL defined in file 'LICENSE' with
# LICENSE_EXCEPTION and the Commercial licence defined in file 'COMM_LICENSE',
# which is part of this source code package.

from django import forms
from viewflow.utils import viewprop
from viewflow.forms import Span


def _collect_elements(parent, container=None):
    if container is None:
        container = []

    if hasattr(parent, 'children'):
        for element in parent.children:
            _collect_elements(element, container=container)

    if isinstance(parent, Span):
        container.append(parent.field_name)

    return container


class FormLayoutMixin(object):
    """
    Mixin for FormView to infer View.fields definition from form Layout.
    """
    form_class = None

    @viewprop
    def layout(self):
        if self.form_class is None and hasattr(self.form_class, 'layout'):
            return self.form_class.layout

    @viewprop
    def fields(self):
        if self.form_class is None:
            if self.layout is not None:
                return _collect_elements(self.layout)
            else:
                return '__all__'


class Action(object):
    def __init__(self, name, url=None, viewname=None, icon=None):
        assert url or viewname
        self.name = name
        self.url = url
        self.viewname = viewname
        self.icon = icon


class BulkActionForm(forms.Form):
    def __init__(self, *args, **kwargs):
        model = kwargs.pop('model')
        super().__init__(*args, **kwargs)

        self.fields['pk'] = forms.ModelMultipleChoiceField(
            queryset=model._default_manager.all(),
            widget=forms.MultipleHiddenInput,
            required=False
        )

        self.fields['select_all'] = forms.CharField(
            widget=forms.HiddenInput,
            required=False
        )



# Copyright (c) 2017-2020, Mikhail Podgurskiy
# All Rights Reserved.

# This work is dual-licensed under AGPL defined in file 'LICENSE' with
# LICENSE_EXCEPTION and the Commercial license defined in file 'COMM_LICENSE',
# which is part of this source code package.

from typing import Callable, Any, List, Union, Mapping, TYPE_CHECKING
from viewflow.this_object import ThisObject

if TYPE_CHECKING:
    from .base import TransitionMethod, Transition  # NOQA

UserModel = Any
StateValue = Any
Condition = Union[ThisObject, Callable[[object, object], bool]]
Permission = Union[ThisObject, Callable[[object, Any], bool]]
StateTransitions = Mapping['TransitionMethod', List['Transition']]
TransitionFunction = Callable[..., Any]



from django.db import models
from django.utils.translation import pgettext_lazy


class PROCESS(models.TextChoices):
    CANCELED = 'CANCELED', pgettext_lazy('STATUS', 'Canceled')
    DONE = 'DONE', pgettext_lazy('STATUS', 'Done')
    NEW = 'NEW', pgettext_lazy('STATUS', 'New')


class STATUS(models.TextChoices):
    ASSIGNED = 'ASSIGNED', pgettext_lazy('STATUS', 'Assigned')
    CANCELED = 'CANCELED', pgettext_lazy('STATUS', 'Canceled')
    DONE = 'DONE', pgettext_lazy('STATUS', 'Done')
    ERROR = 'ERROR', pgettext_lazy('STATUS', 'Error')
    NEW = 'NEW', pgettext_lazy('STATUS', 'New')
    REVIVED = 'REVIVED', pgettext_lazy('STATUS', 'Revived')  # TODO
    SCHEDULED = 'SCHEDULED', pgettext_lazy('STATUS', 'Scheduled')
    STARTED = 'STARTED', pgettext_lazy('STATUS', 'Started')

    def __str__(self):
        return str(self.value)



# Copyright (c) 2017-2020, Mikhail Podgurskiy
# All Rights Reserved.

# This work is licensed under the Commercial license defined in file
# 'COMM_LICENSE', which is part of this source code package.

from contextlib import contextmanager
from functools import partial

from django.conf import settings
from django.contrib.admin import helpers
from django.contrib.admin.utils import flatten_fieldsets, quote
from django.contrib.admin.exceptions import DisallowedModelAdminLookup
from django.core.exceptions import PermissionDenied
from django.forms import Media
from django.forms.models import modelform_factory
from django.http import HttpResponseRedirect
from django.shortcuts import render
from django.urls import path, reverse
from django.utils.translation import gettext_lazy as _

from viewflow import fsm
from viewflow.fsm.base import State, TransitionBoundMethod


class FlowAdminMixin(object):
    """
    A Mixin for providing Finite State Machine (FSM) management support in
    Django admin.

    """

    flow_state: State

    change_list_template = "admin/fsm_change_list.html"
    change_form_template = "admin/fsm_change_form.html"

    transition_form_template = None

    def get_flow_state(self, request) -> State:
        return self.flow_state

    def get_object_flow(self, request, obj):
        """
        Returns the flow object associated with the specified model object.

        This function retrieves the flow object associated with the given model
        object. Override this function if your flow class does not have a
        constructor that accepts the model object as the only argument.
        """
        try:
            return self.get_flow_state()._owner(obj)
        except TypeError:
            raise ValueError(
                f"{self.flow_state._owner} does not have a constructor that accepts"
                " a single argument. Please redefine 'get_object_flow' on the model"
                " admin."
            )

    def get_transition_fields(self, request, obj, slug):
        """
        Override this method to return a list of editable fields for the form
        associated with the specified transition slug.

        If no fields are editable for the transition, this function should return
        None.
        """
        return None

    def save_model(self, request, obj, form, change):
        if not change:
            state = self.get_flow_state(request)
            flow = self.get_object_flow(request, obj)
            state.set(flow, state.get(flow))
        super().save_model(request, obj, form, change)

    @contextmanager
    def create_revision(self, request):
        """
        Save wrapper to use in state transition views, in case of subclass of
        django-reversion VersionAdmin.
        """
        if hasattr(super(), "create_revision"):
            with super().create_revision(request):
                yield
        else:
            yield

    def get_urls(self):
        info = self.model._meta.app_label, self.model._meta.model_name
        return [
            path(
                "<path:object_id>/transition/<slug:slug>/",
                self.admin_site.admin_view(self.transition_view),
                name="%s_%s_transition" % info,
            ),
        ] + super().get_urls()

    @property
    def media(self):
        extra = ".min"  # '' if settings.DEBUG else '.min'

        return super().media + Media(
            css={
                "screen": (
                    "viewflow/css/vis-network%s.css" % extra,
                    "viewflow/css/viewflow%s.css" % extra,
                )
            },
            js=[
                "viewflow/js/vis-network%s.js" % extra,
                "viewflow/js/viewflow%s.js" % extra,
            ],
        )

    def changelist_view(self, request, extra_context=None):
        state = self.get_flow_state(request)
        flow_chart = fsm.chart(state)

        return super().changelist_view(
            request, extra_context={"state": state, "flow_chart": flow_chart}
        )

    def render_change_form(
        self, request, context, add=False, change=False, form_url="", obj=None
    ):
        if change:
            state = self.get_flow_state(request)
            flow = self.get_object_flow(request, obj)

            context.update(
                {
                    "state": state,
                    "flow": flow,
                    "transitions": [
                        (
                            transition,
                            transition.conditions_met(flow),
                            transition.has_perm(flow, request.user),
                        )
                        for transition in state.get_outgoing_transitions(
                            state.get(flow)
                        )
                    ],
                }
            )

        return super().render_change_form(
            request, context, add=add, change=change, form_url=form_url, obj=obj
        )

    def transition_view(self, request, object_id, slug):
        opts = self.model._meta
        obj = self.get_object(request, object_id)
        if obj is None:
            return self._get_obj_does_not_exist_redirect(request, opts, object_id)

        flow = self.get_object_flow(request, obj)
        transition = getattr(flow, slug, None) if not slug.startswith("_") else None
        if not transition or not isinstance(transition, TransitionBoundMethod):
            raise DisallowedModelAdminLookup

        if not transition.has_perm(request.user):
            raise PermissionDenied

        if not transition.can_proceed():
            raise PermissionDenied(_("Transition is not allowed"))

        # build form
        fields = self.get_transition_fields(request, obj, slug) or []
        readonly_fields = [
            field
            for field in flatten_fieldsets(self.get_fieldsets(request, obj))
            if field not in fields
        ]

        ModelForm = modelform_factory(
            self.model,
            **{
                "form": self.form,
                "fields": fields,
                "formfield_callback": partial(
                    self.formfield_for_dbfield, request=request
                ),
            },
        )
        form = ModelForm(request.POST or None, instance=obj)
        adminForm = helpers.AdminForm(
            form,
            list(self.get_fieldsets(request, obj)),
            {},
            readonly_fields,
            model_admin=self,
        )
        media = self.media + adminForm.media

        if form.is_valid():
            # perform transition
            self.save_model(request, obj, form, change=True)
            transition()

            obj_url = reverse(
                "admin:%s_%s_change" % (opts.app_label, opts.model_name),
                args=(quote(obj.pk),),
                current_app=self.admin_site.name,
            )

            return HttpResponseRedirect(obj_url)

        context = {
            **self.admin_site.each_context(request),
            "title": _("%(label)s %(name)s")
            % ({"label": transition.label, "name": opts.verbose_name}),
            "adminform": adminForm,
            "object_id": object_id,
            "original": obj,
            "media": media,
            "preserved_filters": self.get_preserved_filters(request),
            "transition": transition,
            "opts": opts,
            "has_view_permission": self.has_view_permission(request, obj),
        }

        return render(
            request,
            self.transition_form_template
            or [
                "admin/%s/%s/fsm_transition_form.html"
                % (opts.app_label, opts.model_name),
                "admin/%s/fsm_transition_form.html" % opts.app_label,
                "admin/fsm_transition_form.html",
            ],
            context,
        )



from django.db import models


class Device(models.Model):
    created = models.DateTimeField('Создано', auto_now_add=True)
    user_id = models.PositiveIntegerField()
    uuid = models.CharField('UUID', max_length=40, help_text='Уникальный идентификатор устройства')
    ip = models.CharField(max_length=64)
    isp = models.CharField('ISP', max_length=256)
    fingerprint = models.CharField('Fingerprint', max_length=32)
    country = models.CharField('Страна', max_length=4)

    def save(self, *args, **kwargs):
        pass

    class Meta:
        managed = False
        verbose_name = 'Устройство'
        verbose_name_plural = 'Устройства'



from django.urls import path

from viewflow import viewprop, this
from viewflow.workflow import nodes
from . import views, mixins, utils


class Start(mixins.NodeDetailMixin, mixins.NodeUndoMixin, nodes.Start):
    """
    The ``Start`` node in a flow.

    This node is used as the initial step in a flow by a user.

    `Live Demo <https://demo.viewflow.io/workflow/flows/helloworld/start/>`_ /
    `Cookbook sample <https://github.com/viewflow/cookbook/blob/main/workflow101/helloworld/flows.py>`_

    .. code-block:: python

        class MyFlow(flow.Flow):
            start = (
                flow.Start(views.CreateProcessView.as_view(fields=["text"]))
                .Annotation(title=_("New message"))
                .Permission(auto_create=True)
                .Next(this.approve)
            )

            ...

    """

    index_view_class = views.IndexTaskView
    detail_view_class = views.DetailTaskView
    undo_view_class = views.UndoTaskView

    @property
    def start_view(self):
        return this.resolve(self.flow_class.instance, self._start_view)

    @property
    def start_view_path(self):
        return path(
            f"{self.name}/",
            utils.wrap_start_view(self, self.start_view),
            name="execute",
        )


class StartHandle(mixins.NodeDetailMixin, mixins.NodeUndoMixin, nodes.StartHandle):
    """
    The ``Start`` handle node in a flow.

    This node is used as the initial step in a flow from code

    .. code-block:: python

        class MyFlow(flow.Flow):
            start = flow.StartHandle(this.on_start_process).Next(this.approve)

            def start_process(self, activation, sample=False):
                activation.process.sample = sample
                return activation.process

            ...

        process = MyFlow.start.run(sample=True)

    """

    index_view_class = views.IndexTaskView
    detail_view_class = views.DetailTaskView
    undo_view_class = views.UndoTaskView


class End(
    mixins.NodeDetailMixin, mixins.NodeUndoMixin, mixins.NodeReviveMixin, nodes.End
):
    """
    The ``End`` node in a flow.

    This node serves as the terminal point of a flow

    .. code-block:: python

        class MyFlow(flow.Flow):
            ...

            approved = this.End()
            rejected = this.End()

    """

    index_view_class = views.IndexTaskView
    detail_view_class = views.DetailTaskView
    undo_view_class = views.UndoTaskView
    revive_view_class = views.ReviveTaskView


class View(
    mixins.NodeDetailMixin,
    mixins.NodeCancelMixin,
    mixins.NodeUndoMixin,
    mixins.NodeReviveMixin,
    nodes.View,
):
    """
    Represents a user-interaction node within a flow

    .. code-block:: python

        class MyFlow(flow.Flow):
            ...

            approve = (
                flow.View(views.UpdateProcessView.as_view(fields=["approved"]))
                .Annotation(
                    title=_("Approve"),
                    description=_("Supervisor approvement"),
                    summary_template=_("Message review required"),
                    result_template=_(
                        "Message was {{ process.approved|yesno:'Approved,Rejected' }}"
                    ),
                )
                .Permission(auto_create=True)
                .Next(this.check_approve)
            )

            ...
    """

    index_view_class = views.UserIndexTaskView
    detail_view_class = views.DetailTaskView
    cancel_view_class = views.CancelTaskView
    undo_view_class = views.UndoTaskView
    revive_view_class = views.ReviveTaskView

    """
    Execute View
    """

    @property
    def view(self):
        return this.resolve(self.flow_class.instance, self._view)

    @property
    def view_path(self):
        return path(
            f"<int:process_pk>/{self.name}/<int:task_pk>/execute/",
            utils.wrap_view(self, self.view),
            name="execute",
        )

    """
    Assign user to a task
    """
    assign_view_class = views.AssignTaskView

    @viewprop
    def assign_view(self):
        """View for a task assign."""
        if self.assign_view_class:
            return self.assign_view_class.as_view()

    @property
    def assign_path(self):
        if self.assign_view:
            return path(
                f"<int:process_pk>/{self.name}/<int:task_pk>/assign/",
                utils.wrap_task_view(
                    self, self.assign_view, permission=self.can_assign
                ),
                name="assign",
            )

    """
    Unassign
    """
    unassign_view_class = views.UnassignTaskView

    @viewprop
    def unassign_view(self):
        """View for a task assign."""
        if self.unassign_view_class:
            return self.unassign_view_class.as_view()

    @property
    def unassign_path(self):
        if self.unassign_view:
            return path(
                f"<int:process_pk>/{self.name}/<int:task_pk>/unassign/",
                utils.wrap_task_view(
                    self, self.unassign_view, permission=self.can_unassign
                ),
                name="unassign",
            )


class If(
    mixins.NodeDetailMixin, mixins.NodeUndoMixin, mixins.NodeReviveMixin, nodes.If
):
    """
    The  If-gate

    .. code-block:: python

        class MyFlow(flow.Flow):
            ...

            check_approve = (
                flow.If(lambda activation: activation.process.approved)
                .Annotation(title=_("Approvement check"))
                .Then(this.send)
                .Else(this.end)
            )

     .       ...
    """

    index_view_class = views.IndexTaskView
    detail_view_class = views.DetailTaskView
    undo_view_class = views.UndoTaskView
    revive_view_class = views.ReviveTaskView


class Function(mixins.NodeDetailMixin, mixins.NodeUndoMixin, nodes.Function):
    index_view_class = views.IndexTaskView
    detail_view_class = views.DetailTaskView
    undo_view_class = views.UndoTaskView


class Handle(
    mixins.NodeDetailMixin,
    mixins.NodeCancelMixin,
    mixins.NodeUndoMixin,
    mixins.NodeReviveMixin,
    nodes.Handle,
):
    """
    Represents a task executed from the other parts of code

    Usage:
    To define a handle in a flow and run it:

    .. code-block:: python

        class MyFlow(flow.Flow):
            ...
            my_handle = flow.Handle().Next(this.join_gate)

        task = task=process.task_set.get(flow_task=MyFlow.my_handle, status=STATUS.NEW),
        MyFlow.my_handle.run(task)
    """

    index_view_class = views.IndexTaskView
    detail_view_class = views.DetailTaskView
    cancel_view_class = views.CancelTaskView
    undo_view_class = views.UndoTaskView
    revive_view_class = views.ReviveTaskView


class Obsolete(mixins.NodeDetailMixin, mixins.NodeCancelMixin, nodes.Obsolete):
    index_view_class = views.IndexTaskView
    detail_view_class = views.DetailTaskView
    cancel_view_class = views.CancelTaskView


class Join(
    mixins.NodeDetailMixin,
    mixins.NodeExecuteMixin,
    mixins.NodeCancelMixin,
    mixins.NodeUndoMixin,
    mixins.NodeReviveMixin,
    nodes.Join,
):
    index_view_class = views.IndexTaskView
    detail_view_class = views.DetailTaskView
    cancel_view_class = views.CancelTaskView
    undo_view_class = views.UndoTaskView
    revive_view_class = views.ReviveTaskView


class Split(
    mixins.NodeDetailMixin,
    mixins.NodeExecuteMixin,
    mixins.NodeUndoMixin,
    mixins.NodeReviveMixin,
    nodes.Split,
):
    index_view_class = views.IndexTaskView
    detail_view_class = views.DetailTaskView
    undo_view_class = views.UndoTaskView
    revive_view_class = views.ReviveTaskView


class SplitFirst(
    mixins.NodeDetailMixin,
    nodes.SplitFirst,
):
    index_view_class = views.IndexTaskView
    detail_view_class = views.DetailTaskView


class StartSubprocess(
    mixins.NodeDetailMixin,
    mixins.NodeCancelMixin,
    mixins.NodeUndoMixin,
    nodes.StartSubprocess,
):
    index_view_class = views.IndexTaskView
    detail_view_class = views.DetailTaskView
    cancel_view_class = views.CancelTaskView
    undo_view_class = views.UndoTaskView


class Subprocess(
    mixins.NodeDetailMixin,
    mixins.NodeCancelMixin,
    mixins.NodeUndoMixin,
    nodes.Subprocess,
):
    index_view_class = views.IndexTaskView
    detail_view_class = views.DetailTaskView
    cancel_view_class = views.CancelTaskView
    undo_view_class = views.UndoTaskView


class NSubprocess(
    mixins.NodeDetailMixin,
    mixins.NodeCancelMixin,
    mixins.NodeUndoMixin,
    nodes.NSubprocess,
):
    index_view_class = views.IndexTaskView
    detail_view_class = views.DetailTaskView
    cancel_view_class = views.CancelTaskView
    undo_view_class = views.UndoTaskView


class Switch(
    mixins.NodeDetailMixin,
    mixins.NodeCancelMixin,
    mixins.NodeUndoMixin,
    mixins.NodeReviveMixin,
    nodes.Switch,
):
    index_view_class = views.IndexTaskView
    detail_view_class = views.DetailTaskView
    cancel_view_class = views.CancelTaskView
    undo_view_class = views.UndoTaskView
    revive_view_class = views.ReviveTaskView



from django.forms import model_to_dict
from rest_framework import HTTP_HEADER_ENCODING
from rest_framework.pagination import PageNumberPagination
from rest_framework_simplejwt.authentication import JWTAuthentication
from rest_framework_simplejwt.settings import api_settings


def get_client_ip(request):
    """Получить ip клиента из request"""
    x_forwarded_for = request.META.get('HTTP_X_FORWARDED_FOR')
    if x_forwarded_for:
        ip = x_forwarded_for.split(',')[0]
    else:
        ip = request.META.get('REMOTE_ADDR')
    return ip


class CurrentUserIdDefault:
    """Получить user_id из request для сериалайзера"""
    requires_context = True

    def __call__(self, serializer_field):
        return serializer_field.context['request'].user.id

    def __repr__(self):
        return '%s()' % self.__class__.__name__


class ClientIpDefault:
    """Получить ip клиента из request для сериалайзера"""
    requires_context = True

    def __call__(self, serializer_field):
        return get_client_ip(serializer_field.context['request'])

    def __repr__(self):
        return '%s()' % self.__class__.__name__


class PageSizePagination(PageNumberPagination):
    page_size_query_param = 'page_size'


class CustomJWTAuthentication(JWTAuthentication):
    def get_header(self, request):
        """
        Для авторизации с заголовком: Auth: Bearer 8e2928a5f341513728ee6870c57652582a3af99c
        """
        if request.META.get('HTTP_AUTH'):
            header = request.META.get('HTTP_AUTH')
        else:
            header = request.META.get(api_settings.AUTH_HEADER_NAME)
        if isinstance(header, str):
            header = header.encode(HTTP_HEADER_ENCODING)
        return header


class ModelDiffMixin:
    """
    A model mixin that tracks model fields' values and provide some useful api
    to know what fields have been changed.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__initial = self._dict

    @property
    def diff(self):
        d1 = self.__initial
        d2 = self._dict
        diffs = [(k, (v, d2[k])) for k, v in d1.items() if v != d2[k]]
        return dict(diffs)

    @property
    def has_changed(self):
        return bool(self.diff)

    @property
    def changed_fields(self):
        return self.diff.keys()

    def get_field_diff(self, field_name):
        """
        Returns a diff for field if it's changed and None otherwise.
        """
        return self.diff.get(field_name, None)

    def save(self, *args, **kwargs):
        """
        Saves model and set initial state.
        """
        super().save(*args, **kwargs)
        self.__initial = self._dict

    @property
    def _dict(self):
        return model_to_dict(self, fields=[field.name for field in
                                           self._meta.fields])



from django.contrib import admin
from django.contrib.auth.models import Group, User
from rest_framework.authtoken.models import TokenProxy

admin.site.unregister(User)
admin.site.unregister(Group)
admin.site.unregister(TokenProxy)



from django.db import models


class PayInterests(models.Model):
    created = models.DateTimeField('Создана', auto_now_add=True)
    loan = models.ForeignKey('loans.Loan', models.CASCADE, verbose_name='Займ')
    amount = models.DecimalField('Сумма', max_digits=7, decimal_places=0)
    is_export = models.BooleanField('Выгружать', default=True, help_text='Без галочки: не отправлять данные в 1С')
    comment = models.CharField('Комментарий', max_length=1024, blank=True)

    def save(self, *args, **kwargs):
        pass

    def __str__(self):
        return f'{self.pk}'

    class Meta:
        managed = False
        db_table = 'operations_payinterests'
        verbose_name = 'Погашенные проценты'
        verbose_name_plural = 'Погашенные проценты'


class PayBody(models.Model):
    created = models.DateTimeField('Создана', auto_now_add=True)
    loan = models.ForeignKey('loans.Loan', models.CASCADE, verbose_name='Займ')
    amount = models.DecimalField('Сумма', max_digits=7, decimal_places=0)
    is_export = models.BooleanField('Выгружать"', default=True, help_text='Не отправлять данные в 1С')
    comment = models.CharField('Комментарий', max_length=1024, blank=True)

    def save(self, *args, **kwargs):
        pass

    def __str__(self):
        return f'{self.pk}'

    class Meta:
        managed = False
        db_table = 'operations_paybody'
        verbose_name = 'Погашенное тело'
        verbose_name_plural = 'Погашенное тело'


class BodyIssue(models.Model):
    """
    Перечисление денег по займу на карту
    """
    BANK_CARD = 'BANK_CARD'
    SBP = 'SBP'
    METHOD_CHOICES = {
        BANK_CARD: 'Банковская карта',
        SBP: 'СБП'
    }

    created = models.DateTimeField('Создана', auto_now_add=True)
    loan = models.ForeignKey('loans.Loan', on_delete=models.CASCADE, verbose_name='Займ')
    bank_card_id = models.PositiveIntegerField('Банковская карта', help_text='Карта для займа', null=True, blank=True)
    amount = models.DecimalField('Сумма', max_digits=7, decimal_places=0, null=True)
    payment_id = models.CharField('TransactionId', help_text='Если заполнено, значит деньги отправлены', max_length=64)
    is_export = models.BooleanField('Выгружать"', default=True, help_text='Не отправлять данные в 1С')
    method = models.CharField('Метод выплаты/оплаты', max_length=16, choices=sorted(METHOD_CHOICES.items()),
                              default=BANK_CARD)
    sbp_member = models.CharField(max_length=32, blank=True, help_text='идентификатор банка при отправке через СБП')
    comment = models.CharField(max_length=1024, blank=True)

    def save(self, *args, **kwargs):
        pass

    def __str__(self):
        return f'{self.pk}'

    class Meta:
        managed = False
        db_table = 'operations_bodyissue'
        verbose_name = 'Выдача займа'
        verbose_name_plural = 'Выдача займа'



# Copyright (c) 2017-2020, Mikhail Podgurskiy
# All Rights Reserved.

# This work is dual-licensed under AGPL defined in file 'LICENSE' with
# LICENSE_EXCEPTION and the Commercial license defined in file 'COMM_LICENSE',
# which is part of this source code package.

from django.urls import NoReverseMatch
from django.utils.functional import cached_property

from viewflow.utils import Icon, camel_case_to_title, strip_suffixes

from .base import IndexViewMixin, Viewset


class AppMenuMixin:
    """A route that can be listed in an Application menu."""

    title = None
    icon = Icon("view_carousel")

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        if isinstance(self.icon, str):
            self.icon = Icon(self.icon)

    def __getattribute__(self, name):
        attr = super().__getattribute__(name)

        if name == "title" and attr is None:
            class_title = camel_case_to_title(
                strip_suffixes(
                    self.__class__.__name__, ["Viewset", "Admin", "App", "Flow"]
                )
            )
            if not class_title:
                raise ValueError("Application item needs a title")
            return class_title + "s"

        return attr

    def has_view_permission(self, user, obj=None):
        if hasattr(super(), "has_view_permission"):
            return super().has_view_permission(user, obj=obj)
        return True


class Application(IndexViewMixin, Viewset):
    title = ""
    icon = Icon("view_module")
    menu_template_name = "viewflow/includes/app_menu.html"
    base_template_name = "viewflow/base_page.html"
    permission = None

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        if isinstance(self.icon, str):
            self.icon = Icon(self.icon)

    def __getattribute__(self, name):
        attr = super().__getattribute__(name)

        if name == "title" and attr is not None and not attr:
            title = camel_case_to_title(
                strip_suffixes(
                    self.__class__.__name__,
                    ["Application", "Viewset", "Admin", "App", "Flow"],
                )
            )
            if not title:
                raise ValueError("Application needs a title")
            return title

        return attr

    def _get_resolver_extra(self):
        return {"viewset": self, "app": self}

    def get_context_data(self, request):
        return {}

    def has_view_permission(self, user, obj=None):
        if self.permission is not None:
            if callable(self.permission):
                return self.permission(user)
            return user.is_authenticated and user.has_perm(self.permission)
        return True

    def menu_items(self):
        for viewset in self._children:
            if isinstance(viewset, AppMenuMixin):
                yield viewset


class Site(IndexViewMixin, Viewset):
    title = None
    icon = Icon("view_comfy")
    menu_template_name = "viewflow/includes/site_menu.html"
    primary_color = None
    secondary_color = None
    permission = None

    def __init__(self, *, title=None, **kwargs):
        super().__init__(**kwargs)

        if isinstance(self.icon, str):
            self.icon = Icon(self.icon)

        if title is not None:
            self.title = title

        if self.title is None:
            # pluralize class name
            self.title = camel_case_to_title(
                strip_suffixes(self.__class__.__name__, ["Site"])
            )
            if not self.title:
                self.title = "Django Viewflow"

    def __getattribute__(self, name):
        attr = super().__getattribute__(name)

        if name == "title" and attr is None:
            title = camel_case_to_title(
                strip_suffixes(self.__class__.__name__, ["Site"])
            )
            if not title:
                title = "Django Viewflow"
            return title

        return attr

    def _get_resolver_extra(self):
        return {"viewset": self, "site": self}

    def menu_items(self):
        for viewset in self._children:
            if isinstance(viewset, (Site, Application)):
                yield viewset

    def has_view_permission(self, user, obj=None):
        if self.permission is not None:
            return user.has_perm(self.permission)
        return True

    def register(self, app_class):
        app = app_class()
        app._parent = self
        self.viewsets.append(app)
        return app_class

    @cached_property
    def _viewset_models(self):
        result = {}

        queue = list(self._children)
        while queue:
            viewset = queue.pop(0)
            if (
                hasattr(viewset, "model")
                and hasattr(viewset, "get_object_url")
                and viewset.model not in result
            ):
                result[viewset.model] = viewset
            for child_viewset in viewset._children:
                if isinstance(child_viewset, Viewset):
                    queue.append(child_viewset)
        return result

    def get_absolute_url(self, request, obj):
        model = type(obj)
        if model in self._viewset_models:
            return self._viewset_models[model].get_object_url(request, obj)
        else:
            raise NoReverseMatch("Viewset for {} not found".format(model.__name__))



from decimal import Decimal

from django.db import models


class Loan(models.Model):
    ISSUING = 'ISSUING'
    OPEN = "OPEN"  # Клиент получил деньги
    OVERDUE = "OVERDUE"
    STOP_INTEREST = "STOP_INTEREST"
    CLOSED = "CLOSED"
    CANCELED = 'CANCELED'
    STATUS_CHOICES = {
        ISSUING: 'Выдается',
        OPEN: "Открыт",
        OVERDUE: "Просрочен",
        STOP_INTEREST: "Остановлена проценты",
        CLOSED: "Закрыт",
        CANCELED: 'Отменен',
    }

    BANK_CARD = 'BANK_CARD'
    SBP = 'SBP'
    METHOD_CHOICES = {
        BANK_CARD: 'Банковская карта',
        SBP: 'СБП'
    }

    created = models.DateTimeField('Создан', auto_now_add=True)
    updated = models.DateTimeField('Изменен', auto_now=True)
    order = models.OneToOneField('orders.Order', on_delete=models.PROTECT, verbose_name='Заявка')
    extended_start_date = models.DateField('Начало пролонгации', null=True)
    extended_end_date = models.DateField('Окончание пролонгации', null=True)
    bank_card_id = models.PositiveIntegerField('Банковская карта', null=True, help_text='Карта для зачисления займа')
    interests_charged_date = models.DateField('Проценты начислены', null=True, help_text='Дата последнего начисления')
    status = models.CharField('Статус', max_length=16, choices=sorted(STATUS_CHOICES.items()), default=ISSUING)
    expiry_date = models.DateField('Дата окончания', help_text='По договору без учета пролонгации')
    closed_date = models.DateTimeField('Дата закрытия', null=True, help_text='Дата фактического закрытия')
    sign = models.PositiveSmallIntegerField('Подпись')
    ip = models.GenericIPAddressField()
    document = models.FileField('Договор', upload_to="files/", blank=True, help_text='Печатный договор займа')
    comment = models.CharField('Комментарий', max_length=2048, blank=True)

    interest_rate = models.DecimalField('Процентная ставка', max_digits=3, decimal_places=2)
    period = models.PositiveSmallIntegerField('Срок в днях')
    free_period = models.PositiveSmallIntegerField('Беспроцентный период')

    body_issue = models.DecimalField('Выдано тело', max_digits=7, decimal_places=0, default=Decimal(0))
    interests_charged = models.DecimalField('Начислено процентов', max_digits=7, decimal_places=0, default=Decimal(0))
    interests_paid = models.DecimalField('Выплачено процентов', max_digits=7, decimal_places=0, default=Decimal(0))
    body_paid = models.DecimalField('Выплачено тело', max_digits=7, decimal_places=0, default=Decimal(0))
    method = models.CharField('Метод выплаты/оплаты', max_length=16, choices=sorted(METHOD_CHOICES.items()),
                              default=BANK_CARD)

    @property
    def interests_balance(self):
        """Долг по процентам"""
        return self.interests_charged - self.interests_paid

    @property
    def body_balance(self):
        """Долг по телу"""
        return self.body_issue - self.body_paid

    @property
    def balance(self):
        """Общий долг"""
        return self.interests_balance + self.body_balance

    def save(self, *args, **kwargs):
        pass

    def __str__(self):
        return f'{self.pk}'

    class Meta:
        managed = False
        verbose_name_plural = 'Займы'
        verbose_name = 'Займ'



"""Import flows before auth permission setup."""

# Copyright (c) 2017-2020, Mikhail Podgurskiy
# All Rights Reserved.

# This work is dual-licensed under AGPL defined in file 'LICENSE' with
# LICENSE_EXCEPTION and the Commercial licence defined in file 'COMM_LICENSE',
# which is part of this source code package.

import os
import warnings

from django.apps import apps
from django.db import DEFAULT_DB_ALIAS, router
from django.db.models.signals import pre_migrate, post_migrate


def create_permissions(app_config, verbosity=2, interactive=True, using=DEFAULT_DB_ALIAS, **kwargs):
    """Create permissions on django 1.10+."""
    if not app_config.models_module:
        return

    try:
        ContentType = apps.get_model('contenttypes', 'ContentType')
        Permission = apps.get_model('auth', 'Permission')
    except LookupError:
        return

    if not router.allow_migrate_model(using, Permission):
        return

    searched_perms = list()  # (content_type, (codename, name))
    ctypes = set()           # The codenames and ctypes that should exist.

    for klass in app_config.get_models():
        # ctype = ContentType.objects.db_manager(using).get_for_model(klass)
        opts = klass._meta
        ctype, _ = ContentType.objects.get_or_create(
            app_label=opts.app_label,
            model=opts.object_name.lower()
        )
        ctypes.add(ctype)

        from django.contrib.auth.management import _get_all_permissions
        for perm in _get_all_permissions(klass._meta):
            searched_perms.append((ctype, perm))

    all_perms = set(Permission.objects.using(using).filter(
        content_type__in=ctypes,
    ).values_list(
        "content_type", "codename"
    ))

    perms = [
        Permission(codename=codename, name=name, content_type=ct)
        for ct, (codename, name) in searched_perms
        if (ct.pk, codename) not in all_perms
    ]

    Permission.objects.using(using).bulk_create(perms)
    if verbosity >= 2:
        for perm in perms:
            print("Adding permission '%s'" % perm)


def import_flows(app_config, **kwargs):
    """Pre-import flows to allow permissions auto-creation."""
    try:
        __import__('{}.flows'.format(app_config.module.__name__))
    except ImportError as err:
        if os.path.exists(os.path.join(app_config.path, 'flows.py')):
            warnings.warn(
                f"'{err}' during import flows.py for {app_config.name}\n\n"
                "Flow nodes custom permissions might not autocreated",
                stacklevel=2
            )


pre_migrate.connect(import_flows, dispatch_uid="viewflow.management.import_flows")
post_migrate.connect(create_permissions, dispatch_uid="viewflow.management.create_permissions")



# Copyright (c) 2017-2020, Mikhail Podgurskiy
# All Rights Reserved.

# This work is dual-licensed under AGPL defined in file 'LICENSE' with
# LICENSE_EXCEPTION and the Commercial license defined in file 'COMM_LICENSE',
# which is part of this source code package.

import random
from datetime import datetime
from urllib.parse import urljoin

from django import forms
from django.apps import apps
from django.conf import settings
from django.contrib import messages
from django.contrib.auth import views, forms as auth_forms
from django.contrib.auth.decorators import user_passes_test
from django.core.cache import cache
from django.core.cache.utils import make_template_fragment_key
from django.core.files.storage import default_storage
from django.db import models
from django.urls import path
from django.utils.decorators import method_decorator
from django.views import generic
from django.utils.translation import gettext_lazy as _

# from viewflow.forms.widgets import TextInput, PasswordInput
from viewflow.urls import Viewset
from viewflow.utils import viewprop


class AuthenticationForm(auth_forms.AuthenticationForm):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.fields['username'].widget = forms.TextInput(
            attrs={'autofocus': True, 'leading-icon': 'account_box'},
        )
        self.fields['password'].widget = forms.PasswordInput(
            attrs={'leading-icon': 'lock'}
        )


@method_decorator(user_passes_test(lambda u: u.is_authenticated), name='dispatch')
class ProfileView(generic.DetailView):
    template_name = 'registration/profile.html'

    def get_object_data(self):
        for field in self.object._meta.fields:
            if field.name in ['password']:
                continue
            if isinstance(field, models.AutoField):
                continue
            elif field.auto_created:
                continue
            else:
                choice_display_attr = "get_{}_display".format(field.name)
            if hasattr(self.object, choice_display_attr):
                value = getattr(self.object, choice_display_attr)()
            else:
                value = getattr(self.object, field.name)

            if value is not None:
                yield (field, field.verbose_name.capitalize(), value)

    def get_object(self):
        return self.request.user

    def post(self, request, *args, **kwargs):
        class AvatarForm(forms.Form):
            avatar = forms.FileField(required=True)
        form = AvatarForm(request.POST, request.FILES)
        if form.is_valid():
            file_name = 'avatars/{}.png'.format(request.user.pk)
            if default_storage.exists(file_name):
                default_storage.delete(file_name)
            default_storage.save(file_name, form.cleaned_data['avatar'], max_length=512 * 1024)
            key = make_template_fragment_key('django-viewflow-avatar', [request.user.pk])
            cache.delete(key)
            messages.add_message(self.request, messages.SUCCESS, random.choice(GREETINGS), fail_silently=True)
        else:
            message = ''.join(
                f"{field}: " + ''.join(error['message'] for error in errors)
                for field, errors in form.errors.get_json_data(escape_html=True).items()
            )
            messages.add_message(self.request, messages.ERROR, message, fail_silently=True)
        return self.get(request, *args, **kwargs)


def get_user_avatar_url(user):
    """Lookup for user avatar, on the media files."""

    key = make_template_fragment_key('django-viewflow-avatar', [user.pk])
    url = cache.get(key)
    if url is not None and 'LocMemCache' not in settings.CACHES.get('default', {}).get('BACKEND'):
        return url

    file_name = 'avatars/{}.png'.format(user.pk)
    if default_storage.exists(file_name):
        try:
            modified = default_storage.get_modified_time(file_name)
        except NotImplementedError:
            modified = datetime.now()
        url = default_storage.url(file_name) + "?timestamp={}".format(modified.timestamp())
    else:
        if apps.is_installed('django.contrib.staticfiles'):
            from django.contrib.staticfiles.storage import staticfiles_storage
            url = staticfiles_storage.url('viewflow/img/user.png')
        else:
            url = urljoin(settings.STATIC_URL, 'viewflow/img/user.png')

    cache.set(key, url)
    return url


class AuthViewset(Viewset):
    """
    Class-based url configuration for the `django.contrib.auth`.

    urlpatterns = [
        path('accounts/', AuthViewset(
            allow_password_change=False,
            login_view=views.LoginView.as_view(
                authentication_form=MyAuthForm
            ),
        ).urls),
    ]
    """

    def __init__(self, *, allow_password_change=True, with_profile_view=True, **kwargs):
        """
            Initialize the viewset.

            :param allow_password_change=True: enable password change/reset views
        """
        super().__init__(**kwargs)
        self.allow_password_change = allow_password_change
        self.with_profile_view = with_profile_view

    """
    Login
    """
    login_view_class = views.LoginView

    def get_login_view_kwargs(self, **kwargs):
        result = {
            'form_class': AuthenticationForm
        }
        result.update(kwargs)
        return result

    @viewprop
    def login_view(self):
        return self.login_view_class.as_view(**self.get_login_view_kwargs())

    @property
    def login_path(self):
        return path('login/', self.login_view, name='login')

    """
    Logout
    """
    logout_view_class = views.LogoutView

    def get_logout_view_kwargs(self, **kwargs):
        return kwargs

    @viewprop
    def logout_view(self):
        return self.logout_view_class.as_view(**self.get_logout_view_kwargs())

    @property
    def logout_path(self):
        return path('logout/', self.logout_view, name='logout')

    """
    Password Change
    """
    pass_change_view_class = views.PasswordChangeView

    def get_pass_change_view_kwargs(self, **kwargs):
        return kwargs

    @viewprop
    def pass_change_view(self):
        return self.pass_change_view_class.as_view(**self.get_pass_change_view_kwargs())

    @property
    def pass_change_path(self):
        if self.allow_password_change:
            return path(
                'password_change/', self.pass_change_view,
                name='password_change')

    """
    Password Change Done
    """
    pass_change_done_view_class = views.PasswordChangeDoneView

    def get_pass_change_done_view_kwargs(self, **kwargs):
        return kwargs

    @viewprop
    def pass_change_done_view(self):
        return self.pass_change_done_view_class.as_view(**self.get_pass_change_done_view_kwargs())

    @property
    def pass_change_done_path(self):
        if self.allow_password_change:
            return path(
                'password_change/done/', self.pass_change_done_view,
                name='password_change_done')

    """
    Password Reset Request
    """
    pass_reset_view_class = views.PasswordResetView

    def get_pass_reset_view_kwargs(self, **kwargs):
        return kwargs

    @viewprop
    def pass_reset_view(self):
        return self.pass_reset_view_class.as_view(**self.get_pass_reset_view_kwargs())

    @property
    def pass_reset_path(self):
        if self.allow_password_change:
            return path(
                'password_reset/', self.pass_reset_view,
                name='password_reset')

    """
    Password Reset Request Done
    """
    pass_reset_done_view_class = views.PasswordResetDoneView

    def get_pass_reset_done_view_kwargs(self, **kwargs):
        return kwargs

    @viewprop
    def pass_reset_done_view(self):
        return self.pass_reset_done_view_class.as_view(**self.get_pass_reset_done_view_kwargs())

    @property
    def pass_reset_done_path(self):
        if self.allow_password_change:
            return path(
                'password_reset/done/', self.pass_reset_done_view,
                name='password_reset_done')

    """
    Password Reset Request Confirm
    """
    pass_reset_confirm_view_class = views.PasswordResetConfirmView

    def get_pass_reset_confirm_view_kwargs(self, **kwargs):
        return kwargs

    @viewprop
    def pass_reset_confirm_view(self):
        return self.pass_reset_confirm_view_class.as_view(**self.get_pass_reset_confirm_view_kwargs())

    @property
    def pass_reset_confirm_path(self):
        if self.allow_password_change:
            return path(
                'reset/<uidb64>/<token>/', self.pass_reset_confirm_view,
                name='password_reset_confirm')

    """
    Password Request Request Confirmed
    """
    pass_reset_complete_view_class = views.PasswordResetCompleteView

    def get_pass_reset_complete_view_kwargs(self, **kwargs):
        return kwargs

    @viewprop
    def pass_reset_complete_view(self):
        return self.pass_reset_complete_view_class.as_view(**self.get_pass_reset_complete_view_kwargs())

    @property
    def pass_reset_complete_path(self):
        if self.allow_password_change:
            return path(
                'reset/done/', self.pass_reset_complete_view,
                name='password_reset_complete')
    """
    Profile
    """
    profile_view_class = ProfileView

    def get_profile_view_kwargs(self, **kwargs):
        return kwargs

    @viewprop
    def profile_view(self):
        return self.profile_view_class.as_view(**self.get_profile_view_kwargs())

    @property
    def profile_path(self):
        if self.with_profile_view:
            return path('profile/', self.profile_view, name='profile')

    """
    Django-allauth integration
    """
    def get_allauth_prodivers(self):
        try:
            from allauth.socialaccount import providers
            return providers.registry.get_list()
        except ImportError:
            return []


GREETINGS = [
    _('Fantastic!'),
    _('That looks awesome!'),
    _('You are looking very well today!'),
    _('I totally admire your spontaneity.'),
    _('I like your new haircut.'),
    _('What a beautiful costume!'),
    _('You look very good in that suit'),
    _('I love your style.'),
    _('I love your hair today'),
    _('That color looks great on you!'),
]



from django.http import HttpResponseRedirect
from django.views import generic
from django.utils.translation import gettext_lazy as _
from viewflow.views import FormLayoutMixin
from . import mixins


class UpdateProcessView(
    FormLayoutMixin,
    mixins.SuccessMessageMixin,
    mixins.TaskSuccessUrlMixin,
    mixins.TaskViewTemplateNames,
    generic.UpdateView,
):
    """Default view to update a process"""

    success_message = _("Task {task} has been completed.")
    template_filename = "task.html"

    def get_object(self):
        """Return the process for the task activation."""
        return self.request.activation.process

    def form_valid(self, form):
        """If the form is valid, save the associated model and finish the task."""
        self.object = form.save()
        self.request.activation.execute()
        return HttpResponseRedirect(self.get_success_url())


class UpdateArtifactView(
    FormLayoutMixin,
    mixins.SuccessMessageMixin,
    mixins.TaskSuccessUrlMixin,
    mixins.TaskViewTemplateNames,
    generic.UpdateView,
):
    success_message = _("Task {task} has been completed.")
    template_filename = "task.html"

    def get_object(self):
        """Return the process for the task activation."""
        return self.request.activation.process.artifact

    def form_valid(self, form):
        """If the form is valid, save the associated model and finish the task."""
        self.object = form.save()
        self.request.activation.execute()
        return HttpResponseRedirect(self.get_success_url())



import threading


_context_stack = threading.local()


class Context(object):
    """Thread-local activation context, dynamically scoped.

    :keyword propagate_exception: If True, on activation failure
                                  exception will be propagated to
                                  previous activation. If False,
                                  current task activation will be
                                  marked as failed.


    Usage ::

        with Context(propagate_exception=False):
             print(context.propagate_exception)  # prints 'False'
        print(context.propagate_exception)  # prints default 'True'

    """

    def __init__(self, default=None, **kwargs):  # noqa D102
        self.default = default
        self.current_context_data = kwargs

    def __getattr__(self, name):
        stack = []

        if hasattr(_context_stack, 'data'):
            stack = _context_stack.data

        for scope in reversed(stack):
            if name in scope:
                return scope[name]

        if name in self.default:
            return self.default[name]

        raise AttributeError(name)

    def __enter__(self):
        if not hasattr(_context_stack, 'data'):
            _context_stack.data = []
        _context_stack.data.append(self.current_context_data)

    def __exit__(self, t, v, tb):
        _context_stack.data.pop()

    @staticmethod
    def create(**kwargs):  # noqa D102
        return Context(default=kwargs)


context = Context.create(propagate_exception=True)



# Copyright (c) 2017-2020, Mikhail Podgurskiy
# All Rights Reserved.

# This work is dual-licensed under AGPL defined in file 'LICENSE' with
# LICENSE_EXCEPTION and the Commercial license defined in file 'COMM_LICENSE',
# which is part of this source code package.

from django.urls import path

from viewflow.urls import AppMenuMixin, Viewset, ViewsetMeta
from viewflow.utils import DEFAULT, Icon, first_not_default, has_object_perm, viewprop
from viewflow.views import (
    Action,
    CreateModelView,
    DeleteBulkActionView,
    DeleteModelView,
    DetailModelView,
    ListModelView,
    UpdateModelView,
)


class BaseModelViewset(Viewset):
    model = DEFAULT
    queryset = DEFAULT

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        assert self.model is not DEFAULT, "No model specified for {}".format(self)

    def __getattribute__(self, name):
        attr = super(BaseModelViewset, self).__getattribute__(name)
        if name == "title" and attr is None:
            return self.model._meta.verbose_name_plural.capitalize()
        return attr

    def filter_kwargs(self, view_class, **kwargs):
        return super().filter_kwargs(
            view_class,
            **{
                "model": self.model,
                "viewset": self,
                "queryset": self.queryset,
                **kwargs,
            }
        )

    @property
    def index_path(self):
        return path("", self.list_view, name="index")

    """
    List
    """
    list_view_class = ListModelView
    list_columns = DEFAULT
    list_paginate_by = DEFAULT
    list_object_link_columns = DEFAULT
    list_page_actions = DEFAULT
    list_filterset_class = DEFAULT
    list_filter_fields = DEFAULT
    list_search_fields = DEFAULT
    list_ordering_fields = DEFAULT

    def has_view_permission(self, user, obj=None):
        if has_object_perm(user, "view", self.model, obj=obj):
            return True
        if hasattr(self, "has_change_permission"):
            return self.has_change_permission(user, obj=obj)
        return False

    def get_list_page_actions(self, request, *actions):
        return (*self.list_page_actions, *actions)

    def get_list_view_kwargs(self, **kwargs):
        view_kwargs = {
            "columns": self.list_columns,
            "paginate_by": self.list_paginate_by,
            "object_link_columns": self.list_object_link_columns,
            "filterset_class": self.list_filterset_class,
            "filter_fields": self.list_filter_fields,
            "search_fields": self.list_search_fields,
            "ordering": self.list_ordering_fields,
            **self.list_view_kwargs,
            **kwargs,
        }
        return self.filter_kwargs(self.list_view_class, **view_kwargs)

    @viewprop
    def list_view_kwargs(self):
        return {}

    @viewprop
    def list_view(self):
        return self.list_view_class.as_view(**self.get_list_view_kwargs())

    @property
    def list_path(self):
        return path("", self.list_view, name="list")


class ListBulkActionsMixin(metaclass=ViewsetMeta):
    list_bulk_actions = DEFAULT

    def get_list_bulk_actions(self, request, *actions):
        if self.list_bulk_actions is not DEFAULT:
            actions = (*self.list_bulk_actions, *actions)
        return actions


class CreateViewMixin(metaclass=ViewsetMeta):
    create_view_class = CreateModelView
    create_form_layout = DEFAULT
    create_form_class = DEFAULT
    create_form_widgets = DEFAULT

    def has_add_permission(self, user):
        return has_object_perm(user, "add", self.model)

    def get_create_view_kwargs(self, **kwargs):
        view_kwargs = {
            "form_class": first_not_default(
                self.create_form_class, getattr(self, "form_class", DEFAULT)
            ),
            "form_widgets": first_not_default(
                self.create_form_widgets, getattr(self, "form_widgets", DEFAULT)
            ),
            "layout": first_not_default(
                self.create_form_layout, getattr(self, "form_layout", DEFAULT)
            ),
            **self.create_view_kwargs,
            **kwargs,
        }
        return self.filter_kwargs(self.create_view_class, **view_kwargs)

    def get_list_page_actions(self, request, *actions):
        add_action = Action(
            name="Add new",
            url=self.reverse("add"),
            icon=Icon("add_circle", class_="material-icons mdc-list-item__graphic"),
        )
        return super().get_list_page_actions(request, *(add_action, *actions))

    @viewprop
    def create_view_kwargs(self):
        return {}

    @viewprop
    def create_view(self):
        return self.create_view_class.as_view(**self.get_create_view_kwargs())

    @property
    def create_path(self):
        return path("add/", self.create_view, name="add")


class UpdateViewMixin(metaclass=ViewsetMeta):
    update_view_class = UpdateModelView
    update_page_actions = DEFAULT

    update_form_layout = DEFAULT
    update_form_class = DEFAULT
    update_form_widgets = DEFAULT

    form_layout = DEFAULT
    form_class = DEFAULT
    form_widgets = DEFAULT

    def has_change_permission(self, user, obj=None):
        return has_object_perm(user, "change", self.model, obj=obj)

    def get_update_view_kwargs(self, **kwargs):
        view_kwargs = {
            "form_class": first_not_default(self.update_form_class, self.form_class),
            "form_widgets": first_not_default(
                self.update_form_widgets, self.form_widgets
            ),
            "layout": first_not_default(self.update_form_layout, self.form_layout),
            **self.update_view_kwargs,
            **kwargs,
        }
        return self.filter_kwargs(self.update_view_class, **view_kwargs)

    @viewprop
    def update_view_kwargs(self):
        return {}

    @viewprop
    def update_view(self):
        return self.update_view_class.as_view(**self.get_update_view_kwargs())

    @property
    def update_path(self):
        return path("<path:pk>/change/", self.update_view, name="change")

    def get_update_page_actions(self, request, obj, *actions):
        if self.update_page_actions is not DEFAULT:
            actions = (*self.update_page_actions, *actions)
        return actions


class ModelViewset(
    ListBulkActionsMixin,
    CreateViewMixin,
    UpdateViewMixin,
    AppMenuMixin,
    BaseModelViewset,
):
    """List/Create/Update/Delete for a model."""

    def get_object_url(self, request, obj):
        if self.has_change_permission(request.user, obj):
            return self.reverse("change", args=[obj.pk])

    def get_success_url(self, request, obj=None):
        return self.reverse("index")


class DeleteViewMixin(metaclass=ViewsetMeta):
    delete_view_class = DeleteModelView

    def has_delete_permission(self, user, obj=None):
        return has_object_perm(user, "delete", self.model, obj=obj)

    """
    Bulk delete
    """
    bulk_delete_view_class = DeleteBulkActionView

    def get_bulk_delete_view_kwargs(self, **kwargs):
        view_kwargs = {
            "filterset_class": self.list_filterset_class,
            "filter_fields": self.list_filter_fields,
            **self.bulk_delete_view_kwargs,
            **kwargs,
        }
        return self.filter_kwargs(self.bulk_delete_view_class, **view_kwargs)

    @viewprop
    def bulk_delete_view_kwargs(self):
        return {}

    @viewprop
    def bulk_delete_view(self):
        return self.bulk_delete_view_class.as_view(**self.get_bulk_delete_view_kwargs())

    @property
    def bulk_delete_path(self):
        return path("action/delete/", self.bulk_delete_view, name="bulk_delete")

    def get_list_bulk_actions(self, request, *actions):
        if self.has_delete_permission(request.user):
            bulk_delete_action = Action(
                name="Delete selected objects",
                url=self.reverse("bulk_delete"),
                icon=Icon("delete", class_="material-icons mdc-list-item__graphic"),
            )
            actions = (bulk_delete_action, *actions)
        return super().get_list_bulk_actions(request, *actions)

    """
    Delete single object
    """

    def get_delete_view_kwargs(self, **kwargs):
        view_kwargs = {**self.delete_view_kwargs, **kwargs}
        return self.filter_kwargs(self.delete_view_class, **view_kwargs)

    @viewprop
    def delete_view_kwargs(self):
        return {}

    @viewprop
    def delete_view(self):
        return self.delete_view_class.as_view(**self.get_delete_view_kwargs())

    @property
    def delete_path(self):
        return path("<path:pk>/delete/", self.delete_view, name="delete")

    def get_update_page_actions(self, request, obj, *actions):
        if self.has_delete_permission(request.user):
            actions = (
                Action(
                    name="Delete",
                    url=self.reverse("delete", args=[obj.pk]),
                    icon=Icon("delete", class_="material-icons mdc-list-item__graphic"),
                ),
                *actions,
            )
        return super().get_update_page_actions(request, obj, *actions)


class DetailViewMixin(metaclass=ViewsetMeta):
    def get_object_url(self, request, obj):
        if self.has_view_permission(request.user, obj):
            return self.reverse("detail", args=[obj.pk])

    def get_success_url(self, request, obj=None):
        if obj is not None and obj.pk is not None:
            return self.reverse("detail", args=[obj.pk])
        return self.reverse("index")

    """
    Detail
    """
    detail_view_class = DetailModelView
    detail_page_actions = DEFAULT
    detail_page_object_actions = DEFAULT

    def get_detail_view_kwargs(self, **kwargs):
        view_kwargs = {**self.detail_view_kwargs, **kwargs}
        return self.filter_kwargs(self.detail_view_class, **view_kwargs)

    @viewprop
    def detail_view_kwargs(self):
        return {}

    @viewprop
    def detail_view(self):
        return self.detail_view_class.as_view(**self.get_detail_view_kwargs())

    @property
    def detail_path(self):
        return path("<path:pk>/detail/", self.detail_view, name="detail")

    def get_detail_page_actions(self, request, obj, *actions):
        if hasattr(self, "has_delete_permission") and self.has_delete_permission(
            request.user
        ):
            actions = (
                Action(
                    name="Delete",
                    url=self.reverse("delete", args=[obj.pk]),
                    icon=Icon("delete", class_="material-icons mdc-list-item__graphic"),
                ),
                *actions,
            )
        if self.detail_page_actions is not DEFAULT:
            actions = (*self.detail_page_actions, *actions)
        return actions

    def get_detail_page_object_actions(self, request, obj, *actions):
        if self.detail_page_object_actions is not DEFAULT:
            return (*self.detail_page_object_actions, *actions)
        return actions


class ReadonlyModelViewset(
    DetailViewMixin, ListBulkActionsMixin, AppMenuMixin, BaseModelViewset
):
    """
    Readonly model viewset with List and object details view only
    """
